import gc
import threading
import time
import uuid
from typing import Optional

from loguru import logger

from carvekit.api.interface import Interface
from carvekit.web.schemas.config import WebAPIConfig
from carvekit.web.utils.init_utils import init_interface
from carvekit.web.other.removebg import process_remove_bg


class MLProcessor(threading.Thread):
    """Simple ml task queue processor"""
    def __init__(self, api_config: WebAPIConfig):
        super().__init__()
        self.api_config = api_config
        self.interface: Optional[Interface] = None
        self.jobs = {}
        self.completed_jobs = {}

    def run(self):
        """Starts listening for new jobs."""
        unused_completed_jobs_timer = time.time()
        if self.interface is None:
            self.interface = init_interface(self.api_config)
        while True:
            # Clear unused completed jobs every hour
            if time.time() - unused_completed_jobs_timer > 60:
                self.clear_old_completed_jobs()
                unused_completed_jobs_timer = time.time()

            if len(self.jobs.keys()) >= 1:
                id = list(self.jobs.keys())[0]
                data = self.jobs[id]
                # TODO add pydantic scheme here
                response = process_remove_bg(self.interface, data[0], data[1], data[2], data[3])
                self.completed_jobs[id] = [response, time.time()]
                try:
                    del self.jobs[id]
                except KeyError or NameError as e:
                    logger.error(f"Something went wrong with Task Queue: {str(e)}")
                gc.collect()
            else:
                time.sleep(1)
                continue

    def clear_old_completed_jobs(self):
        """Clears old completed jobs"""

        if len(self.completed_jobs.keys()) >= 1:
            for job_id in self.completed_jobs.keys():
                job_finished_time = self.completed_jobs[job_id][1]
                if time.time() - job_finished_time > 3600:
                    try:
                        del self.completed_jobs[job_id]
                    except KeyError or NameError as e:
                        logger.error(f"Something went wrong with Task Queue: {str(e)}")
            gc.collect()

    def job_status(self, id: str) -> str:
        """
        Returns current job status

        Args:
            id: id of the job

        Returns:
            Current job status for specified id. Job status can be [finished, wait, not_found]
        """
        if id in self.completed_jobs.keys():
            return "finished"
        elif id in self.jobs.keys():
            return "wait"
        else:
            return "not_found"

    def job_result(self, id: str):
        """
        Returns job processing result.

        Args:
            id: id of the job

        Returns:
            job processing result.
        """
        if id in self.completed_jobs.keys():
            data = self.completed_jobs[id][0]
            try:
                del self.completed_jobs[id]
            except KeyError or NameError:
                pass
            return data
        else:
            return False

    def job_create(self, data: list):
        """
        Send job to ML Processor

        Args:
            data: data object
        """
        if self.is_alive() is False:
            self.start()
        id = uuid.uuid4().hex
        self.jobs[id] = data
        return id



from pathlib import Path

import click
import tqdm

from carvekit.utils.image_utils import ALLOWED_SUFFIXES
from carvekit.utils.pool_utils import batch_generator, thread_pool_processing
from carvekit.web.schemas.config import MLConfig
from carvekit.web.utils.init_utils import init_interface
from carvekit.utils.fs_utils import save_file


@click.command('removebg', help="Performs background removal on specified photos using console interface.")
@click.option('-i', required=True, type=str, help='Path to input file or dir')
@click.option('-o', default="none", type=str, help="Path to output file or dir")
@click.option('--pre', default='none', type=str, help='Preprocessing method')
@click.option('--post', default='fba', type=str, help='Postprocessing method.')
@click.option('--net', default='u2net', type=str, help='Segmentation Network')
@click.option('--recursive', default=False, type=bool, help='Enables recursive search for images in a folder')
@click.option('--batch_size', default=10, type=int, help='Batch Size for list of images to be loaded to RAM')
@click.option('--batch_size_seg', default=5, type=int,
              help='Batch size for list of images to be processed by segmentation '
                   'network')
@click.option('--batch_size_mat', default=1, type=int, help='Batch size for list of images to be processed by matting '
                                                            'network')
@click.option('--seg_mask_size', default=320, type=int,
              help='The size of the input image for the segmentation neural network.')
@click.option('--matting_mask_size', default=2048, type=int,
              help='The size of the input image for the matting neural network.')
@click.option('--device', default="cpu", type=str,
              help='Processing Device.')
def removebg(i: str, o: str, pre: str, post: str, net: str, recursive: bool,
             batch_size: int, batch_size_seg: int, batch_size_mat: int, seg_mask_size: int,
             matting_mask_size: int,
             device: str):
    out_path = Path(o)
    input_path = Path(i)
    if input_path.is_dir():
        if recursive:
            all_images = input_path.rglob("*.*")
        else:
            all_images = input_path.glob("*.*")
        all_images = [i for i in all_images if i.suffix.lower() in ALLOWED_SUFFIXES
                      and '_bg_removed' not in i.name]
    else:
        all_images = [input_path]

    interface_config = MLConfig(
        segmentation_network=net,
        preprocessing_method=pre,
        postprocessing_method=post,
        device=device,
        batch_size_seg=batch_size_seg,
        batch_size_matting=batch_size_mat,
        seg_mask_size=seg_mask_size,
        matting_mask_size=matting_mask_size
    )

    interface = init_interface(interface_config)

    for image_batch in tqdm.tqdm(batch_generator(all_images, n=batch_size),
                                 total=int(len(all_images) / batch_size),
                                 desc="Removing background", unit=" image batch",
                                 colour="blue"):
        images_without_background = interface(image_batch)  # Remove background
        thread_pool_processing(lambda x: save_file(out_path, image_batch[x], images_without_background[x]),
                               range((len(image_batch))))  # Drop images to fs


if __name__ == "__main__":
    removebg()



"""
Source url: https://github.com/NathanUA/BASNet
Modified by Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: MIT License
"""
import torch
import torch.nn as nn
from torchvision import models


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class BasicBlockDe(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlockDe, self).__init__()

        self.convRes = conv3x3(inplanes, planes, stride)
        self.bnRes = nn.BatchNorm2d(planes)
        self.reluRes = nn.ReLU(inplace=True)

        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = self.convRes(x)
        residual = self.bnRes(residual)
        residual = self.reluRes(residual)

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class RefUnet(nn.Module):
    def __init__(self, in_ch, inc_ch):
        super(RefUnet, self).__init__()

        self.conv0 = nn.Conv2d(in_ch, inc_ch, 3, padding=1)

        self.conv1 = nn.Conv2d(inc_ch, 64, 3, padding=1)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu1 = nn.ReLU(inplace=True)

        self.pool1 = nn.MaxPool2d(2, 2, ceil_mode=True)

        self.conv2 = nn.Conv2d(64, 64, 3, padding=1)
        self.bn2 = nn.BatchNorm2d(64)
        self.relu2 = nn.ReLU(inplace=True)

        self.pool2 = nn.MaxPool2d(2, 2, ceil_mode=True)

        self.conv3 = nn.Conv2d(64, 64, 3, padding=1)
        self.bn3 = nn.BatchNorm2d(64)
        self.relu3 = nn.ReLU(inplace=True)

        self.pool3 = nn.MaxPool2d(2, 2, ceil_mode=True)

        self.conv4 = nn.Conv2d(64, 64, 3, padding=1)
        self.bn4 = nn.BatchNorm2d(64)
        self.relu4 = nn.ReLU(inplace=True)

        self.pool4 = nn.MaxPool2d(2, 2, ceil_mode=True)

        self.conv5 = nn.Conv2d(64, 64, 3, padding=1)
        self.bn5 = nn.BatchNorm2d(64)
        self.relu5 = nn.ReLU(inplace=True)

        self.conv_d4 = nn.Conv2d(128, 64, 3, padding=1)
        self.bn_d4 = nn.BatchNorm2d(64)
        self.relu_d4 = nn.ReLU(inplace=True)

        self.conv_d3 = nn.Conv2d(128, 64, 3, padding=1)
        self.bn_d3 = nn.BatchNorm2d(64)
        self.relu_d3 = nn.ReLU(inplace=True)

        self.conv_d2 = nn.Conv2d(128, 64, 3, padding=1)
        self.bn_d2 = nn.BatchNorm2d(64)
        self.relu_d2 = nn.ReLU(inplace=True)

        self.conv_d1 = nn.Conv2d(128, 64, 3, padding=1)
        self.bn_d1 = nn.BatchNorm2d(64)
        self.relu_d1 = nn.ReLU(inplace=True)

        self.conv_d0 = nn.Conv2d(64, 1, 3, padding=1)

        self.upscore2 = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=False)

    def forward(self, x):
        hx = x
        hx = self.conv0(hx)

        hx1 = self.relu1(self.bn1(self.conv1(hx)))
        hx = self.pool1(hx1)

        hx2 = self.relu2(self.bn2(self.conv2(hx)))
        hx = self.pool2(hx2)

        hx3 = self.relu3(self.bn3(self.conv3(hx)))
        hx = self.pool3(hx3)

        hx4 = self.relu4(self.bn4(self.conv4(hx)))
        hx = self.pool4(hx4)

        hx5 = self.relu5(self.bn5(self.conv5(hx)))

        hx = self.upscore2(hx5)

        d4 = self.relu_d4(self.bn_d4(self.conv_d4(torch.cat((hx, hx4), 1))))
        hx = self.upscore2(d4)

        d3 = self.relu_d3(self.bn_d3(self.conv_d3(torch.cat((hx, hx3), 1))))
        hx = self.upscore2(d3)

        d2 = self.relu_d2(self.bn_d2(self.conv_d2(torch.cat((hx, hx2), 1))))
        hx = self.upscore2(d2)

        d1 = self.relu_d1(self.bn_d1(self.conv_d1(torch.cat((hx, hx1), 1))))

        residual = self.conv_d0(d1)

        return x + residual


class BASNet(nn.Module):
    def __init__(self, n_channels, n_classes):
        super(BASNet, self).__init__()

        resnet = models.resnet34(pretrained=False)

        # -------------Encoder--------------

        self.inconv = nn.Conv2d(n_channels, 64, 3, padding=1)
        self.inbn = nn.BatchNorm2d(64)
        self.inrelu = nn.ReLU(inplace=True)

        # stage 1
        self.encoder1 = resnet.layer1  # 224
        # stage 2
        self.encoder2 = resnet.layer2  # 112
        # stage 3
        self.encoder3 = resnet.layer3  # 56
        # stage 4
        self.encoder4 = resnet.layer4  # 28

        self.pool4 = nn.MaxPool2d(2, 2, ceil_mode=True)

        # stage 5
        self.resb5_1 = BasicBlock(512, 512)
        self.resb5_2 = BasicBlock(512, 512)
        self.resb5_3 = BasicBlock(512, 512)  # 14

        self.pool5 = nn.MaxPool2d(2, 2, ceil_mode=True)

        # stage 6
        self.resb6_1 = BasicBlock(512, 512)
        self.resb6_2 = BasicBlock(512, 512)
        self.resb6_3 = BasicBlock(512, 512)  # 7

        # -------------Bridge--------------

        # stage Bridge
        self.convbg_1 = nn.Conv2d(512, 512, 3, dilation=2, padding=2)  # 7
        self.bnbg_1 = nn.BatchNorm2d(512)
        self.relubg_1 = nn.ReLU(inplace=True)
        self.convbg_m = nn.Conv2d(512, 512, 3, dilation=2, padding=2)
        self.bnbg_m = nn.BatchNorm2d(512)
        self.relubg_m = nn.ReLU(inplace=True)
        self.convbg_2 = nn.Conv2d(512, 512, 3, dilation=2, padding=2)
        self.bnbg_2 = nn.BatchNorm2d(512)
        self.relubg_2 = nn.ReLU(inplace=True)

        # -------------Decoder--------------

        # stage 6d
        self.conv6d_1 = nn.Conv2d(1024, 512, 3, padding=1)  # 16
        self.bn6d_1 = nn.BatchNorm2d(512)
        self.relu6d_1 = nn.ReLU(inplace=True)

        self.conv6d_m = nn.Conv2d(512, 512, 3, dilation=2, padding=2)  ###
        self.bn6d_m = nn.BatchNorm2d(512)
        self.relu6d_m = nn.ReLU(inplace=True)

        self.conv6d_2 = nn.Conv2d(512, 512, 3, dilation=2, padding=2)
        self.bn6d_2 = nn.BatchNorm2d(512)
        self.relu6d_2 = nn.ReLU(inplace=True)

        # stage 5d
        self.conv5d_1 = nn.Conv2d(1024, 512, 3, padding=1)  # 16
        self.bn5d_1 = nn.BatchNorm2d(512)
        self.relu5d_1 = nn.ReLU(inplace=True)

        self.conv5d_m = nn.Conv2d(512, 512, 3, padding=1)  ###
        self.bn5d_m = nn.BatchNorm2d(512)
        self.relu5d_m = nn.ReLU(inplace=True)

        self.conv5d_2 = nn.Conv2d(512, 512, 3, padding=1)
        self.bn5d_2 = nn.BatchNorm2d(512)
        self.relu5d_2 = nn.ReLU(inplace=True)

        # stage 4d
        self.conv4d_1 = nn.Conv2d(1024, 512, 3, padding=1)  # 32
        self.bn4d_1 = nn.BatchNorm2d(512)
        self.relu4d_1 = nn.ReLU(inplace=True)

        self.conv4d_m = nn.Conv2d(512, 512, 3, padding=1)  ###
        self.bn4d_m = nn.BatchNorm2d(512)
        self.relu4d_m = nn.ReLU(inplace=True)

        self.conv4d_2 = nn.Conv2d(512, 256, 3, padding=1)
        self.bn4d_2 = nn.BatchNorm2d(256)
        self.relu4d_2 = nn.ReLU(inplace=True)

        # stage 3d
        self.conv3d_1 = nn.Conv2d(512, 256, 3, padding=1)  # 64
        self.bn3d_1 = nn.BatchNorm2d(256)
        self.relu3d_1 = nn.ReLU(inplace=True)

        self.conv3d_m = nn.Conv2d(256, 256, 3, padding=1)  ###
        self.bn3d_m = nn.BatchNorm2d(256)
        self.relu3d_m = nn.ReLU(inplace=True)

        self.conv3d_2 = nn.Conv2d(256, 128, 3, padding=1)
        self.bn3d_2 = nn.BatchNorm2d(128)
        self.relu3d_2 = nn.ReLU(inplace=True)

        # stage 2d

        self.conv2d_1 = nn.Conv2d(256, 128, 3, padding=1)  # 128
        self.bn2d_1 = nn.BatchNorm2d(128)
        self.relu2d_1 = nn.ReLU(inplace=True)

        self.conv2d_m = nn.Conv2d(128, 128, 3, padding=1)  ###
        self.bn2d_m = nn.BatchNorm2d(128)
        self.relu2d_m = nn.ReLU(inplace=True)

        self.conv2d_2 = nn.Conv2d(128, 64, 3, padding=1)
        self.bn2d_2 = nn.BatchNorm2d(64)
        self.relu2d_2 = nn.ReLU(inplace=True)

        # stage 1d
        self.conv1d_1 = nn.Conv2d(128, 64, 3, padding=1)  # 256
        self.bn1d_1 = nn.BatchNorm2d(64)
        self.relu1d_1 = nn.ReLU(inplace=True)

        self.conv1d_m = nn.Conv2d(64, 64, 3, padding=1)  ###
        self.bn1d_m = nn.BatchNorm2d(64)
        self.relu1d_m = nn.ReLU(inplace=True)

        self.conv1d_2 = nn.Conv2d(64, 64, 3, padding=1)
        self.bn1d_2 = nn.BatchNorm2d(64)
        self.relu1d_2 = nn.ReLU(inplace=True)

        # -------------Bilinear Upsampling--------------
        self.upscore6 = nn.Upsample(scale_factor=32, mode='bilinear', align_corners=False)  ###
        self.upscore5 = nn.Upsample(scale_factor=16, mode='bilinear', align_corners=False)
        self.upscore4 = nn.Upsample(scale_factor=8, mode='bilinear', align_corners=False)
        self.upscore3 = nn.Upsample(scale_factor=4, mode='bilinear', align_corners=False)
        self.upscore2 = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=False)

        # -------------Side Output--------------
        self.outconvb = nn.Conv2d(512, 1, 3, padding=1)
        self.outconv6 = nn.Conv2d(512, 1, 3, padding=1)
        self.outconv5 = nn.Conv2d(512, 1, 3, padding=1)
        self.outconv4 = nn.Conv2d(256, 1, 3, padding=1)
        self.outconv3 = nn.Conv2d(128, 1, 3, padding=1)
        self.outconv2 = nn.Conv2d(64, 1, 3, padding=1)
        self.outconv1 = nn.Conv2d(64, 1, 3, padding=1)

        # -------------Refine Module-------------
        self.refunet = RefUnet(1, 64)

    def forward(self, x):
        hx = x

        # -------------Encoder-------------
        hx = self.inconv(hx)
        hx = self.inbn(hx)
        hx = self.inrelu(hx)

        h1 = self.encoder1(hx)  # 256
        h2 = self.encoder2(h1)  # 128
        h3 = self.encoder3(h2)  # 64
        h4 = self.encoder4(h3)  # 32

        hx = self.pool4(h4)  # 16

        hx = self.resb5_1(hx)
        hx = self.resb5_2(hx)
        h5 = self.resb5_3(hx)

        hx = self.pool5(h5)  # 8

        hx = self.resb6_1(hx)
        hx = self.resb6_2(hx)
        h6 = self.resb6_3(hx)

        # -------------Bridge-------------
        hx = self.relubg_1(self.bnbg_1(self.convbg_1(h6)))  # 8
        hx = self.relubg_m(self.bnbg_m(self.convbg_m(hx)))
        hbg = self.relubg_2(self.bnbg_2(self.convbg_2(hx)))

        # -------------Decoder-------------

        hx = self.relu6d_1(self.bn6d_1(self.conv6d_1(torch.cat((hbg, h6), 1))))
        hx = self.relu6d_m(self.bn6d_m(self.conv6d_m(hx)))
        hd6 = self.relu6d_2(self.bn6d_2(self.conv6d_2(hx)))

        hx = self.upscore2(hd6)  # 8 -> 16

        hx = self.relu5d_1(self.bn5d_1(self.conv5d_1(torch.cat((hx, h5), 1))))
        hx = self.relu5d_m(self.bn5d_m(self.conv5d_m(hx)))
        hd5 = self.relu5d_2(self.bn5d_2(self.conv5d_2(hx)))

        hx = self.upscore2(hd5)  # 16 -> 32

        hx = self.relu4d_1(self.bn4d_1(self.conv4d_1(torch.cat((hx, h4), 1))))
        hx = self.relu4d_m(self.bn4d_m(self.conv4d_m(hx)))
        hd4 = self.relu4d_2(self.bn4d_2(self.conv4d_2(hx)))

        hx = self.upscore2(hd4)  # 32 -> 64

        hx = self.relu3d_1(self.bn3d_1(self.conv3d_1(torch.cat((hx, h3), 1))))
        hx = self.relu3d_m(self.bn3d_m(self.conv3d_m(hx)))
        hd3 = self.relu3d_2(self.bn3d_2(self.conv3d_2(hx)))

        hx = self.upscore2(hd3)  # 64 -> 128

        hx = self.relu2d_1(self.bn2d_1(self.conv2d_1(torch.cat((hx, h2), 1))))
        hx = self.relu2d_m(self.bn2d_m(self.conv2d_m(hx)))
        hd2 = self.relu2d_2(self.bn2d_2(self.conv2d_2(hx)))

        hx = self.upscore2(hd2)  # 128 -> 256

        hx = self.relu1d_1(self.bn1d_1(self.conv1d_1(torch.cat((hx, h1), 1))))
        hx = self.relu1d_m(self.bn1d_m(self.conv1d_m(hx)))
        hd1 = self.relu1d_2(self.bn1d_2(self.conv1d_2(hx)))

        # -------------Side Output-------------
        db = self.outconvb(hbg)
        db = self.upscore6(db)  # 8->256

        d6 = self.outconv6(hd6)
        d6 = self.upscore6(d6)  # 8->256

        d5 = self.outconv5(hd5)
        d5 = self.upscore5(d5)  # 16->256

        d4 = self.outconv4(hd4)
        d4 = self.upscore4(d4)  # 32->256

        d3 = self.outconv3(hd3)
        d3 = self.upscore3(d3)  # 64->256

        d2 = self.outconv2(hd2)
        d2 = self.upscore2(d2)  # 128->256

        d1 = self.outconv1(hd1)  # 256

        # -------------Refine Module-------------
        dout = self.refunet(d1)  # 256

        return torch.sigmoid(dout), torch.sigmoid(d1), torch.sigmoid(d2), torch.sigmoid(d3), torch.sigmoid(
            d4), torch.sigmoid(d5), torch.sigmoid(
            d6), torch.sigmoid(db)



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import pathlib
from typing import Union, List

import PIL
import numpy as np
import torch
from PIL import Image

from carvekit.ml.arch.basnet.basnet import BASNet
from carvekit.ml.files.models_loc import basnet_pretrained
from carvekit.utils.image_utils import convert_image, load_image
from carvekit.utils.pool_utils import batch_generator, thread_pool_processing

__all__ = ["BASNET"]


class BASNET(BASNet):
    """BASNet model interface"""

    def __init__(self, device='cpu',
                 input_tensor_size: Union[List[int], int] = 320,
                 batch_size: int = 10,
                 load_pretrained: bool = True):
        """
            Initialize the BASNET model

            Args:
                device: processing device
                input_tensor_size: input image size
                batch_size: the number of images that the neural network processes in one run
                load_pretrained: loading pretrained model

        """
        super(BASNET, self).__init__(n_channels=3, n_classes=1)
        self.device = device
        self.batch_size = batch_size
        if isinstance(input_tensor_size, list):
            self.input_image_size = input_tensor_size[:2]
        else:
            self.input_image_size = (input_tensor_size, input_tensor_size)
        self.to(device)
        if load_pretrained:
            self.load_state_dict(torch.load(basnet_pretrained(), map_location=self.device))
        self.eval()

    def data_preprocessing(self, data: PIL.Image.Image) -> torch.Tensor:
        """
            Transform input image to suitable data format for neural network

            Args:
                data: input image

            Returns:
                input for neural network

        """
        resized = data.resize(self.input_image_size)
        # noinspection PyTypeChecker
        resized_arr = np.array(resized, dtype=np.float64)
        temp_image = np.zeros((resized_arr.shape[0], resized_arr.shape[1], 3))
        if np.max(resized_arr) != 0:
            resized_arr /= np.max(resized_arr)
        temp_image[:, :, 0] = (resized_arr[:, :, 0] - 0.485) / 0.229
        temp_image[:, :, 1] = (resized_arr[:, :, 1] - 0.456) / 0.224
        temp_image[:, :, 2] = (resized_arr[:, :, 2] - 0.406) / 0.225
        temp_image = temp_image.transpose((2, 0, 1))
        temp_image = np.expand_dims(temp_image, 0)
        return torch.from_numpy(temp_image).type(torch.FloatTensor)

    @staticmethod
    def data_postprocessing(data: torch.tensor,
                            original_image: PIL.Image.Image) -> PIL.Image.Image:
        """
            Transforms output data from neural network to suitable data
            format for using with other components of this framework.

            Args:
                data: output data from neural network
                original_image: input image which was used for predicted data

            Returns:
                Segmentation mask as PIL Image instance

        """
        data = data.unsqueeze(0)
        mask = data[:, 0, :, :]
        ma = torch.max(mask)  # Normalizes prediction
        mi = torch.min(mask)
        predict = ((mask - mi) / (ma - mi)).squeeze()
        predict_np = predict.cpu().data.numpy() * 255
        mask = Image.fromarray(predict_np).convert("L")
        mask = mask.resize(original_image.size, resample=3)
        return mask

    def __call__(self, images: List[Union[str, pathlib.Path, PIL.Image.Image]]) -> List[PIL.Image.Image]:
        """
            Passes input images through neural network and returns segmentation masks as PIL.Image.Image instances

            Args:
                images: input images

            Returns:
                segmentation masks as for input images, as PIL.Image.Image instances

        """
        collect_masks = []
        for image_batch in batch_generator(images, self.batch_size):
            images = thread_pool_processing(lambda x: convert_image(load_image(x)), image_batch)
            batches = torch.vstack(thread_pool_processing(self.data_preprocessing, images))
            with torch.no_grad():
                batches = batches.to(self.device)
                masks, d2, d3, d4, d5, d6, d7, d8 = super(BASNET, self).__call__(batches)
                masks_cpu = masks.cpu()
                del d2, d3, d4, d5, d6, d7, d8, batches, masks
            masks = thread_pool_processing(lambda x: self.data_postprocessing(masks_cpu[x], images[x]),
                                           range(len(images)))
            collect_masks += masks
        return collect_masks



from pathlib import Path

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.staticfiles import StaticFiles

from carvekit import version
from carvekit.web.deps import config
from carvekit.web.routers.api_router import api_router

app = FastAPI(title='CarveKit Web API', version=version)

app.add_middleware(
    CORSMiddleware,
    allow_origins=['*'],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(api_router, prefix="/api")
app.mount('/', StaticFiles(directory=Path(__file__).parent.joinpath('static'), html=True), name="static")

if __name__ == "__main__":
    uvicorn.run(app, host=config.host, port=config.port)



import secrets
from typing import List
from typing_extensions import Literal

import torch.cuda
from pydantic import BaseModel, validator


class AuthConfig(BaseModel):
    """Config for web api token authentication """
    auth: bool = True
    """Enables Token Authentication for API"""
    admin_token: str = secrets.token_hex(32)
    """Admin Token"""
    allowed_tokens: List[str] = [secrets.token_hex(32)]
    """All allowed tokens"""


class MLConfig(BaseModel):
    """Config for ml part of framework"""
    segmentation_network: Literal["u2net", "deeplabv3", "basnet"] = "u2net"
    """Segmentation Network"""
    preprocessing_method: Literal["none", "stub"] = "none"
    """Pre-processing Method"""
    postprocessing_method: Literal["fba", "none"] = "fba"
    """Post-Processing Network"""
    device: str = "cpu"
    """Processing device"""
    batch_size_seg: int = 5
    """Batch size for segmentation network"""
    batch_size_matting: int = 1
    """Batch size for matting network"""
    seg_mask_size: int = 320
    """The size of the input image for the segmentation neural network."""
    matting_mask_size: int = 2048
    """The size of the input image for the matting neural network."""

    @validator('seg_mask_size')
    def seg_mask_size_validator(cls, value: int, values):
        if value > 0:
            return value
        else:
            raise ValueError("Incorrect seg_mask_size!")

    @validator('matting_mask_size')
    def matting_mask_size_validator(cls, value: int, values):
        if value > 0:
            return value
        else:
            raise ValueError("Incorrect matting_mask_size!")

    @validator('batch_size_seg')
    def batch_size_seg_validator(cls, value: int, values):
        if value > 0:
            return value
        else:
            raise ValueError("Incorrect batch size!")

    @validator('batch_size_matting')
    def batch_size_matting_validator(cls, value: int,values):
        if value > 0:
            return value
        else:
            raise ValueError("Incorrect batch size!")

    @validator('device')
    def device_validator(cls, value):
        if torch.cuda.is_available() is False and "cuda" in value:
            raise ValueError("GPU is not available, but specified as processing device!")
        if 'cuda' not in value and "cpu" != value:
            raise ValueError("Unknown processing device! It should be cpu or cuda!")
        return value


class WebAPIConfig(BaseModel):
    """FastAPI app config"""
    port: int = 5000
    """Web API port"""
    host: str = "0.0.0.0"
    """Web API host"""
    ml: MLConfig = MLConfig()
    """Config for ml part of framework"""
    auth: AuthConfig = AuthConfig()
    """Config for web api token authentication """



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import pathlib
from carvekit.ml.files import checkpoints_dir
from carvekit.utils.download_models import check_for_exists


def u2net_full_pretrained() -> pathlib.Path:
    """ Returns u2net pretrained model location

    Returns:
        pathlib.Path to model location
    """
    return check_for_exists(checkpoints_dir.joinpath('u2net/u2net.pth'))


def basnet_pretrained() -> pathlib.Path:
    """ Returns basnet pretrained model location

    Returns:
        pathlib.Path to model location
    """
    return check_for_exists(checkpoints_dir.joinpath('basnet/basnet.pth'))


def deeplab_pretrained() -> pathlib.Path:
    """ Returns basnet pretrained model location

    Returns:
        pathlib.Path to model location
    """
    return check_for_exists(checkpoints_dir.joinpath('deeplab/deeplab.pth'))


def fba_pretrained() -> pathlib.Path:
    """ Returns basnet pretrained model location

    Returns:
        pathlib.Path to model location
    """
    return check_for_exists(checkpoints_dir.joinpath('fba_matting/fba_matting.pth'))


def download_all():
    u2net_full_pretrained()
    fba_pretrained()
    deeplab_pretrained()
    basnet_pretrained()



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import pathlib
from typing import Union, List, Tuple

import PIL
import cv2
import numpy as np
import torch
from PIL import Image

from carvekit.ml.arch.fba_matting.models import FBA
from carvekit.ml.arch.fba_matting.transforms import trimap_transform, groupnorm_normalise_image
from carvekit.ml.files.models_loc import fba_pretrained
from carvekit.utils.image_utils import convert_image, load_image
from carvekit.utils.pool_utils import batch_generator, thread_pool_processing

__all__ = ["FBAMatting"]


class FBAMatting(FBA):
    """
    FBA Matting Neural Network to improve edges on image.
    """

    def __init__(self, device='cpu',
                 input_tensor_size: Union[List[int], int] = 2048,
                 batch_size: int = 2,
                 encoder="resnet50_GN_WS",
                 load_pretrained: bool = True):
        """
            Initialize the FBAMatting model

            Args:
                device: processing device
                input_tensor_size: input image size
                batch_size: the number of images that the neural network processes in one run
                encoder: neural network encoder head
                load_pretrained: loading pretrained model

        """
        super(FBAMatting, self).__init__(encoder=encoder)
        self.device = device
        self.batch_size = batch_size
        if isinstance(input_tensor_size, list):
            self.input_image_size = input_tensor_size[:2]
        else:
            self.input_image_size = (input_tensor_size, input_tensor_size)
        self.to(device)
        if load_pretrained:
            self.load_state_dict(torch.load(fba_pretrained(), map_location=self.device))
        self.eval()

    def data_preprocessing(self, data: Union[PIL.Image.Image, np.ndarray]) -> Tuple[torch.FloatTensor,
                                                                                    torch.FloatTensor]:
        """
            Transform input image to suitable data format for neural network

            Args:
                data: input image

            Returns:
                input for neural network

        """
        resized = data.copy()
        if self.batch_size == 1:
            resized.thumbnail(self.input_image_size, resample=3)
        else:
            resized = resized.resize(self.input_image_size, resample=3)
        # noinspection PyTypeChecker
        image = np.array(resized, dtype=np.float64)
        image = image / 255.0  # Normalize image to [0, 1] values range
        if resized.mode == 'RGB':
            image = image[:, :, ::-1]
        elif resized.mode == "L":
            image2 = np.copy(image)
            h, w = image2.shape
            image = np.zeros((h, w, 2))  # Transform trimap to binary data format
            image[image2 == 1, 1] = 1
            image[image2 == 0, 0] = 1
        else:
            raise ValueError('Incorrect color mode for image')
        h, w = image.shape[:2]  # Scale input mlt to 8
        h1 = int(np.ceil(1.0 * h / 8) * 8)
        w1 = int(np.ceil(1.0 * w / 8) * 8)
        x_scale = cv2.resize(image, (w1, h1), interpolation=cv2.INTER_LANCZOS4)
        image_tensor = torch.from_numpy(x_scale).permute(2, 0, 1)[None, :, :, :].float()
        if resized.mode == "RGB":
            return image_tensor, groupnorm_normalise_image(image_tensor.clone(), format='nchw')
        else:
            return image_tensor, torch.from_numpy(trimap_transform(x_scale)).permute(2, 0, 1)[None, :, :, :].float()

    @staticmethod
    def data_postprocessing(data: torch.tensor,
                            trimap: PIL.Image.Image) -> PIL.Image.Image:
        """
            Transforms output data from neural network to suitable data
            format for using with other components of this framework.

            Args:
                data: output data from neural network
                trimap: Map with the area we need to refine

            Returns:
                Segmentation mask as PIL Image instance

        """
        if trimap.mode != "L":
            raise ValueError("Incorrect color mode for trimap")
        pred = data.numpy().transpose((1, 2, 0))
        pred = cv2.resize(pred,
                          trimap.size,
                          cv2.INTER_LANCZOS4)[:, :, 0]
        # noinspection PyTypeChecker
        # Clean mask by removing all false predictions outside trimap and already known area
        trimap_arr = np.array(trimap.copy())
        pred[trimap_arr[:, :] == 0] = 0
        # pred[trimap_arr[:, :] == 255] = 1
        pred[pred < 0.3] = 0
        return Image.fromarray(pred * 255).convert("L")

    def __call__(self,
                 images: List[Union[str, pathlib.Path, PIL.Image.Image]],
                 trimaps: List[Union[str, pathlib.Path, PIL.Image.Image]]) -> List[PIL.Image.Image]:
        """
            Passes input images though neural network and returns segmentation masks as PIL.Image.Image instances

            Args:
                images: input images
                trimaps: Maps with the areas we need to refine

            Returns:
                segmentation masks as for input images, as PIL.Image.Image instances

        """

        if len(images) != len(trimaps):
            raise ValueError("Len of specified arrays of images and trimaps should be equal!")

        collect_masks = []
        for idx_batch in batch_generator(range(len(images)), self.batch_size):
            inpt_images = thread_pool_processing(lambda x: convert_image(load_image(images[x])),
                                                 idx_batch)

            inpt_trimaps = thread_pool_processing(lambda x: convert_image(load_image(trimaps[x]), mode="L"),
                                                  idx_batch)

            inpt_img_batches = thread_pool_processing(self.data_preprocessing, inpt_images)
            inpt_trimaps_batches = thread_pool_processing(self.data_preprocessing, inpt_trimaps)

            inpt_img_batches_transformed = torch.vstack([i[1] for i in inpt_img_batches])
            inpt_img_batches = torch.vstack([i[0] for i in inpt_img_batches])

            inpt_trimaps_transformed = torch.vstack([i[1] for i in inpt_trimaps_batches])
            inpt_trimaps_batches = torch.vstack([i[0] for i in inpt_trimaps_batches])

            with torch.no_grad():
                inpt_img_batches = inpt_img_batches.to(self.device)
                inpt_trimaps_batches = inpt_trimaps_batches.to(self.device)
                inpt_img_batches_transformed = inpt_img_batches_transformed.to(self.device)
                inpt_trimaps_transformed = inpt_trimaps_transformed.to(self.device)

                output = super(FBAMatting, self).__call__(inpt_img_batches, inpt_trimaps_batches,
                                                          inpt_img_batches_transformed, inpt_trimaps_transformed)
                output_cpu = output.cpu()
                del inpt_img_batches, inpt_trimaps_batches, \
                    inpt_img_batches_transformed, inpt_trimaps_transformed, output
            masks = thread_pool_processing(lambda x: self.data_postprocessing(output_cpu[x], inpt_trimaps[x]),
                                           range(len(inpt_images)))
            collect_masks += masks
        return collect_masks



"""
Modified by Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
Source url: https://github.com/MarcoForte/FBA_Matting
License: MIT License
"""
import cv2
import numpy as np

group_norm_std = [0.229, 0.224, 0.225]
group_norm_mean = [0.485, 0.456, 0.406]


def dt(a):
    return cv2.distanceTransform((a * 255).astype(np.uint8), cv2.DIST_L2, 0)


def trimap_transform(trimap):
    h, w = trimap.shape[0], trimap.shape[1]

    clicks = np.zeros((h, w, 6))
    for k in range(2):
        if (np.count_nonzero(trimap[:, :, k]) > 0):
            dt_mask = -dt(1 - trimap[:, :, k]) ** 2
            L = 320
            clicks[:, :, 3 * k] = np.exp(dt_mask / (2 * ((0.02 * L) ** 2)))
            clicks[:, :, 3 * k + 1] = np.exp(dt_mask / (2 * ((0.08 * L) ** 2)))
            clicks[:, :, 3 * k + 2] = np.exp(dt_mask / (2 * ((0.16 * L) ** 2)))

    return clicks


def groupnorm_normalise_image(img, format='nhwc'):
    '''
        Accept rgb in range 0,1
    '''
    if (format == 'nhwc'):
        for i in range(3):
            img[..., i] = (img[..., i] - group_norm_mean[i]) / group_norm_std[i]
    else:
        for i in range(3):
            img[..., i, :, :] = (img[..., i, :, :] - group_norm_mean[i]) / group_norm_std[i]

    return img



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import pathlib
from typing import List, Union

import PIL.Image
import torch
from PIL import Image
from torchvision import transforms
from torchvision.models.segmentation import deeplabv3_resnet101
from carvekit.ml.files.models_loc import deeplab_pretrained
from carvekit.utils.image_utils import convert_image, load_image
from carvekit.utils.pool_utils import batch_generator, thread_pool_processing

__all__ = ["DeepLabV3"]


class DeepLabV3:
    def __init__(self, device='cpu',
                 batch_size: int = 10,
                 input_image_size: Union[List[int], int] = 512,
                 load_pretrained: bool = True):
        """
            Initialize the DeepLabV3 model

            Args:
                device: processing device
                input_tensor_size: input image size
                batch_size: the number of images that the neural network processes in one run
                load_pretrained: loading pretrained model

        """
        self.device = device
        self.batch_size = batch_size
        self.network = deeplabv3_resnet101(pretrained=False, pretrained_backbone=False, aux_loss=True)
        self.network.to(self.device)
        if load_pretrained:
            self.network.load_state_dict(torch.load(deeplab_pretrained(), map_location=self.device))
        if isinstance(input_image_size, list):
            self.input_image_size = input_image_size[:2]
        else:
            self.input_image_size = (input_image_size, input_image_size)
        self.network.eval()
        self.data_preprocessing = transforms.Compose([
            transforms.ToTensor(),
            transforms.Resize(self.input_image_size),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ])

    def to(self, device: str):
        """
        Moves neural network to specified processing device

        Args:
            device (:class:`torch.device`): the desired device.
        Returns:
            None

        """
        self.network.to(device)

    @staticmethod
    def data_postprocessing(data: torch.tensor,
                            original_image: PIL.Image.Image) -> PIL.Image.Image:
        """
            Transforms output data from neural network to suitable data
            format for using with other components of this framework.

            Args:
                data: output data from neural network
                original_image: input image which was used for predicted data

            Returns:
                Segmentation mask as PIL Image instance

        """
        return Image.fromarray(data.numpy() * 255).convert("L").resize(original_image.size)

    def __call__(self, images: List[Union[str, pathlib.Path, PIL.Image.Image]]) -> List[PIL.Image.Image]:
        """
            Passes input images though neural network and returns segmentation masks as PIL.Image.Image instances

            Args:
                images: input images

            Returns:
                segmentation masks as for input images, as PIL.Image.Image instances

        """
        collect_masks = []
        for image_batch in batch_generator(images, self.batch_size):
            images = thread_pool_processing(lambda x: convert_image(load_image(x)), image_batch)
            batches = thread_pool_processing(self.data_preprocessing, images)
            with torch.no_grad():
                masks = [self.network(i.to(self.device).unsqueeze(0))['out'][0].argmax(0).byte().cpu() for i in batches]
                del batches
            masks = thread_pool_processing(lambda x: self.data_postprocessing(masks[x], images[x]),
                                           range(len(images)))
            collect_masks += masks
        return collect_masks



"""
    Source url: https://github.com/OPHoperHPO/image-background-remove-tool
    Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
    License: Apache License 2.0
"""

import pathlib
from typing import Union, Any, Tuple

import PIL.Image
import numpy as np
import torch

ALLOWED_SUFFIXES = [".jpg", ".jpeg", ".bmp", ".png", ".webp"]


def to_tensor(x: Any) -> torch.Tensor:
    """
    Returns a PIL.Image.Image as torch tensor without swap tensor dims.

    Args:
        x: PIL.Image.Image instance

    Returns:
        torch.Tensor instance
    """
    return torch.tensor(np.array(x, copy=True))


def load_image(file: Union[str, pathlib.Path, PIL.Image.Image]) -> PIL.Image.Image:
    """ Returns a PIL.Image.Image class by string path or pathlib path or PIL.Image.Image instance

    Args:
        file: File path or PIL.Image.Image instance

    Returns:
        PIL.Image.Image instance

    Raises:
        ValueError: If file not exists or file is directory or file isn't an image or file is not correct PIL Image

    """
    if isinstance(file, str) and is_image_valid(pathlib.Path(file)):
        return PIL.Image.open(file)
    elif isinstance(file, PIL.Image.Image):
        return file
    elif isinstance(file, pathlib.Path) and is_image_valid(file):
        return PIL.Image.open(str(file))
    else:
        raise ValueError("Unknown input file type")


def convert_image(image: PIL.Image.Image, mode="RGB") -> PIL.Image.Image:
    """ Performs image conversion to correct color mode

    Args:
        image: PIL.Image.Image instance
        mode: Colort Mode to convert

    Returns:
        PIL.Image.Image instance

    Raises:
        ValueError: If image hasn't convertable color mode, or it is too small
    """
    if is_image_valid(image):
        return image.convert(mode)


def is_image_valid(image: Union[pathlib.Path, PIL.Image.Image]) -> bool:
    """This function performs image validation.

    Args:
        image: Path to the image or PIL.Image.Image instance being checked.

    Returns:
        True if image is valid

    Raises:
        ValueError: If file not a valid image path or image hasn't convertable color mode, or it is too small

    """
    if isinstance(image, pathlib.Path):
        if not image.exists():
            raise ValueError("File is not exists")
        elif image.is_dir():
            raise ValueError("File is a directory")
        elif image.suffix.lower() not in ALLOWED_SUFFIXES:
            raise ValueError(f"Unsupported image format. Supported file formats: {', '.join(ALLOWED_SUFFIXES)}")
    elif isinstance(image, PIL.Image.Image):
        if not (image.size[0] > 32 and image.size[1] > 32):
            raise ValueError("Image should be bigger then (32x32) pixels.")
        elif image.mode not in ["RGB", "RGBA", "L"]:
            raise ValueError('Wrong image color mode.')
    else:
        raise ValueError("Unknown input file type")
    return True


def transparency_paste(bg_img: PIL.Image.Image, fg_img: PIL.Image.Image, box=(0, 0)) -> PIL.Image.Image:
    """
    Inserts an image into another image while maintaining transparency.

    Args:
        bg_img: background image
        fg_img: foreground image
        box: place to paste

    Returns:
        Background image with pasted foreground image at point or in the specified box
    """
    fg_img_trans = PIL.Image.new("RGBA", bg_img.size)
    fg_img_trans.paste(fg_img, box, mask=fg_img)
    new_img = PIL.Image.alpha_composite(bg_img, fg_img_trans)
    return new_img


def add_margin(pil_img: PIL.Image.Image,
               top: int, right: int, bottom: int, left: int,
               color: Tuple[int, int, int, int])->PIL.Image.Image:
    """
    Adds margin to the image.

    Args:
        pil_img: Image that needed to add margin.
        top: pixels count at top side
        right: pixels count at right side
        bottom: pixels count at bottom side
        left: pixels count at left side
        color: color of margin

    Returns:
        Image with margin.
    """
    width, height = pil_img.size
    new_width = width + right + left
    new_height = height + top + bottom
    # noinspection PyTypeChecker
    result = PIL.Image.new(pil_img.mode, (new_width, new_height), color)
    result.paste(pil_img, (left, top))
    return result



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Iterable


def thread_pool_processing(func: Any, data: Iterable, workers=18):
    """
        Passes all iterator data through the given function

        Args:
            workers: Count of workers.
            func: function to pass data through
            data: input iterator

        Returns:
            function return list

    """
    with ThreadPoolExecutor(workers) as p:
        return list(p.map(func, data))


def batch_generator(iterable, n=1):
    """
        Splits any iterable into n-size packets

        Args:
            iterable: iterator
            n: size of packets

        Returns:
            new n-size packet
    """
    it = len(iterable)
    for ndx in range(0, it, n):
        yield iterable[ndx:min(ndx + n, it)]



from typing import Union

from fastapi import Header
from fastapi.responses import Response, JSONResponse
from carvekit.web.deps import config


def Authenticate(x_api_key: Union[str, None] = Header(None)) -> Union[bool, str]:
    if x_api_key in config.auth.allowed_tokens:
        return "allowed"
    elif x_api_key == config.auth.admin_token:
        return "admin"
    elif config.auth.auth is False:
        return "allowed"
    else:
        return False


def handle_response(response, original_image) -> Response:
    """
    Response handler from TaskQueue
    :param response: TaskQueue response
    :param original_image: Original PIL image
    :return: Complete flask response
    """
    response_object = None
    if isinstance(response, dict):
        if response["type"] == "jpg":
            response_object = Response(content=response["data"][0].read(), media_type='image/jpeg')
        elif response["type"] == "png":
            response_object = Response(content=response["data"][0].read(), media_type='image/png')
        elif response["type"] == "zip":
            response_object = Response(content=response["data"][0], media_type='application/zip')
            response_object.headers['Content-Disposition'] = 'attachment; filename=\'no-bg.zip\''

        # Add headers to output result
        response_object.headers["X-Credits-Charged"] = '0'
        response_object.headers["X-Type"] = "other"  # TODO Make support for this
        response_object.headers["X-Max-Width"] = str(original_image.size[0])
        response_object.headers["X-Max-Height"] = str(original_image.size[1])
        response_object.headers["X-Ratelimit-Limit"] = '500'  # TODO Make ratelimit support
        response_object.headers["X-Ratelimit-Remaining"] = '500'
        response_object.headers["X-Ratelimit-Reset"] = '1'
        response_object.headers["X-Width"] = str(response["data"][1][0])
        response_object.headers["X-Height"] = str(response["data"][1][1])

    else:
        response = JSONResponse(content=response[0])
        response.headers["X-Credits-Charged"] = "0"

    return response_object



import io
import time
import zipfile

import requests
from PIL import Image, ImageColor

from carvekit.utils.image_utils import transparency_paste, add_margin
from carvekit.utils.mask_utils import extract_alpha_channel
from carvekit.web.responses.api import error_dict
from carvekit.api.interface import Interface


def process_remove_bg(interface: Interface, params, image, bg, is_json_or_www_encoded=False):
    """
    Handles a request to the removebg api method

    Args:
        interface: CarveKit interface
        bg: background pil image
        is_json_or_www_encoded: is "json" or "x-www-form-urlencoded" content-type
        image: foreground pil image
        params: parameters
    """
    h, w = image.size
    if h < 2 or w < 2:
        return error_dict("Image is too small. Minimum size 2x2"), 400

    if "size" in params.keys():
        value = params["size"]
        if value == "preview" or value == "small" or value == "regular":
            image.thumbnail((625, 400), resample=3)  # 0.25 mp
        elif value == "medium":
            image.thumbnail((1504, 1000), resample=3)  # 1.5 mp
        elif value == "hd":
            image.thumbnail((2000, 2000), resample=3)  # 2.5 mp
        else:
            image.thumbnail((6250, 4000), resample=3)  # 25 mp

    roi_box = [0, 0, image.size[0], image.size[1]]
    if "type" in params.keys():
        value = params["type"]
        pass

    if "roi" in params.keys():
        value = params["roi"].split(" ")
        if len(value) == 4:
            for i, coord in enumerate(value):
                if "px" in coord:
                    coord = coord.replace("px", "")
                    try:
                        coord = int(coord)
                    except BaseException:
                        return error_dict("Error converting roi coordinate string to number!"), 400
                    if coord < 0:
                        error_dict(
                            "Bad roi coordinate."), 400
                    if (i == 0 or i == 2) and coord > image.size[0]:
                        return error_dict(
                            "The roi coordinate cannot be larger than the image size."), 400
                    elif (i == 1 or i == 3) and coord > image.size[1]:
                        return error_dict(
                            "The roi coordinate cannot be larger than the image size."), 400
                    roi_box[i] = int(coord)
                elif "%" in coord:
                    coord = coord.replace("%", "")
                    try:
                        coord = int(coord)
                    except BaseException:
                        return error_dict("Error converting roi coordinate string to number!"), 400
                    if coord > 100:
                        return error_dict("The coordinate cannot be more than 100%"), 400
                    elif coord < 0:
                        return error_dict("Coordinate cannot be less than 0%"), 400
                    if i == 0 or i == 2:
                        coord = int(image.size[0] * coord / 100)
                    elif i == 1 or i == 3:
                        coord = int(image.size[1] * coord / 100)
                    roi_box[i] = coord
                else:
                    return error_dict("Something wrong with roi coordinates!"), 400

    new_image = image.copy()
    new_image = new_image.crop(roi_box)
    h, w = new_image.size
    if h < 2 or w < 2:
        return error_dict("Image is too small. Minimum size 2x2"), 400
    new_image = interface([new_image])[0]

    scaled = False
    if "scale" in params.keys() and params['scale'] != 100:
        value = params["scale"]
        new_image.thumbnail((int(image.size[0] * value / 100),
                             int(image.size[1] * value / 100)), resample=3)
        scaled = True
    if "crop" in params.keys():
        value = params["crop"]
        if value:
            new_image = new_image.crop(new_image.getbbox())
            if "crop_margin" in params.keys():
                crop_margin = params["crop_margin"]
                if "px" in crop_margin:
                    crop_margin = crop_margin.replace("px", "")
                    crop_margin = abs(int(crop_margin))
                    if crop_margin > 500:
                        return error_dict(
                            "The crop_margin cannot be larger than the original image size."), 400
                    new_image = add_margin(new_image, crop_margin,
                                           crop_margin, crop_margin, crop_margin, (0, 0, 0, 0))
                elif "%" in crop_margin:
                    crop_margin = crop_margin.replace("%", "")
                    crop_margin = int(crop_margin)
                    new_image = add_margin(new_image, int(new_image.size[1] * crop_margin / 100),
                                           int(new_image.size[0] * crop_margin / 100),
                                           int(new_image.size[1] * crop_margin / 100),
                                           int(new_image.size[0] * crop_margin / 100), (0, 0, 0, 0))
        else:
            if "position" in params.keys() and scaled is False:
                value = params["position"]
                if len(value) == 2:
                    new_image = transparency_paste(Image.new("RGBA", image.size), new_image,
                                                   (int(image.size[0] * value[0] / 100),
                                                    int(image.size[1] * value[1] / 100)))
                else:
                    new_image = transparency_paste(Image.new("RGBA", image.size), new_image, roi_box)
            elif scaled is False:
                new_image = transparency_paste(Image.new("RGBA", image.size), new_image, roi_box)

    if "channels" in params.keys():
        value = params["channels"]
        if value == "alpha":
            new_image = extract_alpha_channel(new_image)
        else:
            bg_changed = False
            if "bg_color" in params.keys():
                value = params["bg_color"]
                if len(value) > 0:
                    color = ImageColor.getcolor(value, "RGB")
                    bg = Image.new("RGBA", new_image.size, color)
                    bg = transparency_paste(bg, new_image, (0, 0))
                    new_image = bg.copy()
                    bg_changed = True
            if "bg_image_url" in params.keys() and bg_changed is False:
                value = params["bg_image_url"]
                if len(value) > 0:
                    try:
                        bg = Image.open(io.BytesIO(requests.get(value).content))
                    except BaseException:
                        return error_dict("Error download background image!"), 400
                    bg = bg.resize(new_image.size)
                    bg = bg.convert("RGBA")
                    bg = transparency_paste(bg, new_image, (0, 0))
                    new_image = bg.copy()
                    bg_changed = True
            if not is_json_or_www_encoded:
                if bg and bg_changed is False:
                    bg = bg.resize(new_image.size)
                    bg = bg.convert("RGBA")
                    bg = transparency_paste(bg, new_image, (0, 0))
                    new_image = bg.copy()
    if "format" in params.keys():
        value = params["format"]
        if value == "jpg":
            new_image = new_image.convert("RGB")
            img_io = io.BytesIO()
            new_image.save(img_io, 'JPEG', quality=100)
            img_io.seek(0)
            return {"type": "jpg", "data": [img_io, new_image.size]}
        elif value == "zip":
            mask = extract_alpha_channel(new_image)
            mask_buff = io.BytesIO()
            mask.save(mask_buff, 'PNG')
            mask_buff.seek(0)
            image_buff = io.BytesIO()
            image.save(image_buff, 'JPEG')
            image_buff.seek(0)
            fileobj = io.BytesIO()
            with zipfile.ZipFile(fileobj, 'w') as zip_file:
                zip_info = zipfile.ZipInfo(filename="color.jpg")
                zip_info.date_time = time.localtime(time.time())[:6]
                zip_info.compress_type = zipfile.ZIP_DEFLATED
                zip_file.writestr(zip_info, image_buff.getvalue())
                zip_info = zipfile.ZipInfo(filename="alpha.png")
                zip_info.date_time = time.localtime(time.time())[:6]
                zip_info.compress_type = zipfile.ZIP_DEFLATED
                zip_file.writestr(zip_info, mask_buff.getvalue())
            fileobj.seek(0)
            return {"type": "zip", "data": [fileobj.read(), new_image.size]}
        else:
            buff = io.BytesIO()
            new_image.save(buff, 'PNG')
            buff.seek(0)
            return {"type": "png", "data": [buff, new_image.size]}
    return error_dict("Something wrong with request or http api. Please, open new issue on Github! This is error in "
                      "code."), 400



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import PIL.Image
import torch
from carvekit.utils.image_utils import to_tensor


def composite(foreground: PIL.Image.Image,
              background: PIL.Image.Image,
              alpha: PIL.Image.Image,
              device="cpu"):
    """
    Composites foreground with background by following
    https://pymatting.github.io/intro.html#alpha-matting math formula.

    Args:
        device: Processing device
        foreground: Image that will be pasted to background image with following alpha mask.
        background: Background image
        alpha: Alpha Image

    Returns:
        Composited image as PIL.Image instance.
    """

    foreground = foreground.convert("RGBA")
    background = background.convert("RGBA")
    alpha_rgba = alpha.convert("RGBA")
    alpha_l = alpha.convert("L")

    fg = to_tensor(foreground).to(device)
    alpha_rgba = to_tensor(alpha_rgba).to(device)
    alpha_l = to_tensor(alpha_l).to(device)
    bg = to_tensor(background).to(device)

    alpha_l = alpha_l / 255
    alpha_rgba = alpha_rgba / 255

    bg = torch.where(torch.logical_not(alpha_rgba >= 1), bg, fg)
    bg[:, :, 0] = alpha_l[:, :] * fg[:, :, 0] + (1 - alpha_l[:, :]) * bg[:, :, 0]
    bg[:, :, 1] = alpha_l[:, :] * fg[:, :, 1] + (1 - alpha_l[:, :]) * bg[:, :, 1]
    bg[:, :, 2] = alpha_l[:, :] * fg[:, :, 2] + (1 - alpha_l[:, :]) * bg[:, :, 2]
    bg[:, :, 3] = alpha_l[:, :] * 255

    del alpha_l, alpha_rgba, fg
    return PIL.Image.fromarray(bg.cpu().numpy()).convert("RGBA")


def apply_mask(image: PIL.Image.Image, mask: PIL.Image.Image, device="cpu") -> PIL.Image.Image:
    """
    Applies mask to foreground.

    Args:
        device: Processing device.
        image: Image with background.
        mask: Alpha Channel mask for this image.

    Returns:
        Image without background, where mask was black.
    """
    background = PIL.Image.new("RGBA", image.size, color=(130, 130, 130, 0))
    return composite(image, background, mask, device=device).convert("RGBA")


def extract_alpha_channel(image: PIL.Image.Image) -> PIL.Image.Image:
    """
    Extracts alpha channel from the RGBA image.

    Args:
        image: RGBA PIL image

    Returns:
        RGBA alpha channel image
    """
    alpha = image.split()[-1]
    bg = PIL.Image.new("RGBA", image.size, (0, 0, 0, 255))
    bg.paste(alpha, mask=alpha)
    return bg.convert("RGBA")



"""
Modified by Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
Source url: https://github.com/MarcoForte/FBA_Matting
License: MIT License
"""
import torch
import torch.nn as nn
import carvekit.ml.arch.fba_matting.resnet_GN_WS as resnet_GN_WS
import carvekit.ml.arch.fba_matting.layers_WS as L
import carvekit.ml.arch.fba_matting.resnet_bn as resnet_bn
from functools import partial


class FBA(nn.Module):
    def __init__(self, encoder: str):
        super(FBA, self).__init__()
        self.encoder = build_encoder(arch=encoder)
        self.decoder = fba_decoder(batch_norm=True if 'BN' in encoder else False)

    def forward(self, image, two_chan_trimap, image_n, trimap_transformed):
        resnet_input = torch.cat((image_n, trimap_transformed, two_chan_trimap), 1)
        conv_out, indices = self.encoder(resnet_input, return_feature_maps=True)
        return self.decoder(conv_out, image, indices, two_chan_trimap)


class ResnetDilatedBN(nn.Module):
    def __init__(self, orig_resnet, dilate_scale=8):
        super(ResnetDilatedBN, self).__init__()

        if dilate_scale == 8:
            orig_resnet.layer3.apply(
                partial(self._nostride_dilate, dilate=2))
            orig_resnet.layer4.apply(
                partial(self._nostride_dilate, dilate=4))
        elif dilate_scale == 16:
            orig_resnet.layer4.apply(
                partial(self._nostride_dilate, dilate=2))

        # take pretrained resnet, except AvgPool and FC
        self.conv1 = orig_resnet.conv1
        self.bn1 = orig_resnet.bn1
        self.relu1 = orig_resnet.relu1
        self.conv2 = orig_resnet.conv2
        self.bn2 = orig_resnet.bn2
        self.relu2 = orig_resnet.relu2
        self.conv3 = orig_resnet.conv3
        self.bn3 = orig_resnet.bn3
        self.relu3 = orig_resnet.relu3
        self.maxpool = orig_resnet.maxpool
        self.layer1 = orig_resnet.layer1
        self.layer2 = orig_resnet.layer2
        self.layer3 = orig_resnet.layer3
        self.layer4 = orig_resnet.layer4

    def _nostride_dilate(self, m, dilate):
        classname = m.__class__.__name__
        if classname.find('Conv') != -1:
            # the convolution with stride
            if m.stride == (2, 2):
                m.stride = (1, 1)
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate // 2, dilate // 2)
                    m.padding = (dilate // 2, dilate // 2)
            # other convoluions
            else:
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate, dilate)
                    m.padding = (dilate, dilate)

    def forward(self, x, return_feature_maps=False):
        conv_out = [x]
        x = self.relu1(self.bn1(self.conv1(x)))
        x = self.relu2(self.bn2(self.conv2(x)))
        x = self.relu3(self.bn3(self.conv3(x)))
        conv_out.append(x)
        x, indices = self.maxpool(x)
        x = self.layer1(x)
        conv_out.append(x)
        x = self.layer2(x)
        conv_out.append(x)
        x = self.layer3(x)
        conv_out.append(x)
        x = self.layer4(x)
        conv_out.append(x)

        if return_feature_maps:
            return conv_out, indices
        return [x]


class Resnet(nn.Module):
    def __init__(self, orig_resnet):
        super(Resnet, self).__init__()

        # take pretrained resnet, except AvgPool and FC
        self.conv1 = orig_resnet.conv1
        self.bn1 = orig_resnet.bn1
        self.relu1 = orig_resnet.relu1
        self.conv2 = orig_resnet.conv2
        self.bn2 = orig_resnet.bn2
        self.relu2 = orig_resnet.relu2
        self.conv3 = orig_resnet.conv3
        self.bn3 = orig_resnet.bn3
        self.relu3 = orig_resnet.relu3
        self.maxpool = orig_resnet.maxpool
        self.layer1 = orig_resnet.layer1
        self.layer2 = orig_resnet.layer2
        self.layer3 = orig_resnet.layer3
        self.layer4 = orig_resnet.layer4

    def forward(self, x, return_feature_maps=False):
        conv_out = []

        x = self.relu1(self.bn1(self.conv1(x)))
        x = self.relu2(self.bn2(self.conv2(x)))
        x = self.relu3(self.bn3(self.conv3(x)))
        conv_out.append(x)
        x, indices = self.maxpool(x)

        x = self.layer1(x)
        conv_out.append(x)
        x = self.layer2(x)
        conv_out.append(x)
        x = self.layer3(x)
        conv_out.append(x)
        x = self.layer4(x)
        conv_out.append(x)

        if return_feature_maps:
            return conv_out
        return [x]


class ResnetDilated(nn.Module):
    def __init__(self, orig_resnet, dilate_scale=8):
        super(ResnetDilated, self).__init__()

        if dilate_scale == 8:
            orig_resnet.layer3.apply(
                partial(self._nostride_dilate, dilate=2))
            orig_resnet.layer4.apply(
                partial(self._nostride_dilate, dilate=4))
        elif dilate_scale == 16:
            orig_resnet.layer4.apply(
                partial(self._nostride_dilate, dilate=2))

        # take pretrained resnet, except AvgPool and FC
        self.conv1 = orig_resnet.conv1
        self.bn1 = orig_resnet.bn1
        self.relu = orig_resnet.relu
        self.maxpool = orig_resnet.maxpool
        self.layer1 = orig_resnet.layer1
        self.layer2 = orig_resnet.layer2
        self.layer3 = orig_resnet.layer3
        self.layer4 = orig_resnet.layer4

    def _nostride_dilate(self, m, dilate):
        classname = m.__class__.__name__
        if classname.find('Conv') != -1:
            # the convolution with stride
            if m.stride == (2, 2):
                m.stride = (1, 1)
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate // 2, dilate // 2)
                    m.padding = (dilate // 2, dilate // 2)
            # other convoluions
            else:
                if m.kernel_size == (3, 3):
                    m.dilation = (dilate, dilate)
                    m.padding = (dilate, dilate)

    def forward(self, x, return_feature_maps=False):
        conv_out = [x]
        x = self.relu(self.bn1(self.conv1(x)))
        conv_out.append(x)
        x, indices = self.maxpool(x)
        x = self.layer1(x)
        conv_out.append(x)
        x = self.layer2(x)
        conv_out.append(x)
        x = self.layer3(x)
        conv_out.append(x)
        x = self.layer4(x)
        conv_out.append(x)

        if return_feature_maps:
            return conv_out, indices
        return [x]


def norm(dim, bn=False):
    if bn is False:
        return nn.GroupNorm(32, dim)
    else:
        return nn.BatchNorm2d(dim)


def fba_fusion(alpha, img, F, B):
    F = (alpha * img + (1 - alpha ** 2) * F - alpha * (1 - alpha) * B)
    B = ((1 - alpha) * img + (2 * alpha - alpha ** 2) * B - alpha * (1 - alpha) * F)

    F = torch.clamp(F, 0, 1)
    B = torch.clamp(B, 0, 1)
    la = 0.1
    alpha = (alpha * la + torch.sum((img - B) * (F - B), 1, keepdim=True)) / (
            torch.sum((F - B) * (F - B), 1, keepdim=True) + la)
    alpha = torch.clamp(alpha, 0, 1)
    return alpha, F, B


class fba_decoder(nn.Module):
    def __init__(self, batch_norm=False):
        super(fba_decoder, self).__init__()
        pool_scales = (1, 2, 3, 6)
        self.batch_norm = batch_norm

        self.ppm = []

        for scale in pool_scales:
            self.ppm.append(nn.Sequential(
                nn.AdaptiveAvgPool2d(scale),
                L.Conv2d(2048, 256, kernel_size=1, bias=True),
                norm(256, self.batch_norm),
                nn.LeakyReLU()
            ))
        self.ppm = nn.ModuleList(self.ppm)

        self.conv_up1 = nn.Sequential(
            L.Conv2d(2048 + len(pool_scales) * 256, 256,
                     kernel_size=3, padding=1, bias=True),

            norm(256, self.batch_norm),
            nn.LeakyReLU(),
            L.Conv2d(256, 256, kernel_size=3, padding=1),
            norm(256, self.batch_norm),
            nn.LeakyReLU()
        )

        self.conv_up2 = nn.Sequential(
            L.Conv2d(256 + 256, 256,
                     kernel_size=3, padding=1, bias=True),
            norm(256, self.batch_norm),
            nn.LeakyReLU()
        )
        if self.batch_norm:
            d_up3 = 128
        else:
            d_up3 = 64
        self.conv_up3 = nn.Sequential(
            L.Conv2d(256 + d_up3, 64,
                     kernel_size=3, padding=1, bias=True),
            norm(64, self.batch_norm),
            nn.LeakyReLU()
        )

        self.unpool = nn.MaxUnpool2d(2, stride=2)

        self.conv_up4 = nn.Sequential(
            nn.Conv2d(64 + 3 + 3 + 2, 32,
                      kernel_size=3, padding=1, bias=True),
            nn.LeakyReLU(),
            nn.Conv2d(32, 16,
                      kernel_size=3, padding=1, bias=True),

            nn.LeakyReLU(),
            nn.Conv2d(16, 7, kernel_size=1, padding=0, bias=True)
        )

    def forward(self, conv_out, img, indices, two_chan_trimap):
        conv5 = conv_out[-1]

        input_size = conv5.size()
        ppm_out = [conv5]
        for pool_scale in self.ppm:
            ppm_out.append(nn.functional.interpolate(
                pool_scale(conv5),
                (input_size[2], input_size[3]),
                mode='bilinear', align_corners=False))
        ppm_out = torch.cat(ppm_out, 1)
        x = self.conv_up1(ppm_out)

        x = torch.nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=False)

        x = torch.cat((x, conv_out[-4]), 1)

        x = self.conv_up2(x)
        x = torch.nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=False)

        x = torch.cat((x, conv_out[-5]), 1)
        x = self.conv_up3(x)

        x = torch.nn.functional.interpolate(x, scale_factor=2, mode='bilinear', align_corners=False)
        x = torch.cat((x, conv_out[-6][:, :3], img, two_chan_trimap), 1)

        output = self.conv_up4(x)

        alpha = torch.clamp(output[:, 0][:, None], 0, 1)
        F = torch.sigmoid(output[:, 1:4])
        B = torch.sigmoid(output[:, 4:7])

        # FBA Fusion
        alpha, F, B = fba_fusion(alpha, img, F, B)

        output = torch.cat((alpha, F, B), 1)

        return output


def build_encoder(arch='resnet50_GN'):
    if arch == 'resnet50_GN_WS':
        orig_resnet = resnet_GN_WS.__dict__['l_resnet50']()
        net_encoder = ResnetDilated(orig_resnet, dilate_scale=8)
    elif arch == 'resnet50_BN':
        orig_resnet = resnet_bn.__dict__['l_resnet50']()
        net_encoder = ResnetDilatedBN(orig_resnet, dilate_scale=8)

    else:
        raise ValueError('Architecture undefined!')

    num_channels = 3 + 6 + 2

    if num_channels > 3:
        net_encoder_sd = net_encoder.state_dict()
        conv1_weights = net_encoder_sd['conv1.weight']

        c_out, c_in, h, w = conv1_weights.size()
        conv1_mod = torch.zeros(c_out, num_channels, h, w)
        conv1_mod[:, :3, :, :] = conv1_weights

        conv1 = net_encoder.conv1
        conv1.in_channels = num_channels
        conv1.weight = torch.nn.Parameter(conv1_mod)

        net_encoder.conv1 = conv1

        net_encoder_sd['conv1.weight'] = conv1_mod

        net_encoder.load_state_dict(net_encoder_sd)
    return net_encoder



from carvekit.web.schemas.config import WebAPIConfig
from carvekit.web.utils.init_utils import init_config
from carvekit.web.utils.task_queue import MLProcessor

config: WebAPIConfig = init_config()
ml_processor = MLProcessor(api_config=config)




"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import PIL.Image
import cv2
import numpy as np


class CV2TrimapGenerator:
    def __init__(self, kernel_size: int = 30, erosion_iters: int = 1):
        """
        Initialize a new CV2TrimapGenerator instance

        Args:
            kernel_size: The size of the offset from the object mask
            in pixels when an unknown area is detected in the trimap
            erosion_iters: The number of iterations of erosion that
            the object's mask will be subjected to before forming an unknown area
        """
        self.kernel_size = kernel_size
        self.erosion_iters = erosion_iters

    def __call__(self, original_image: PIL.Image.Image, mask: PIL.Image.Image) -> PIL.Image.Image:
        """
        Generates trimap based on predicted object mask to refine object mask borders.
        Based on cv2 erosion algorithm.

        Args:
            original_image: Original image
            mask: Predicted object mask

        Returns:
            Generated trimap for image.
        """
        if mask.mode != "L":
            raise ValueError("Input mask has wrong color mode.")
        if mask.size != original_image.size:
            raise ValueError("Sizes of input image and predicted mask doesn't equal")
        # noinspection PyTypeChecker
        mask_array = np.array(mask)
        pixels = 2 * self.kernel_size + 1
        kernel = np.ones((pixels, pixels), np.uint8)

        if self.erosion_iters > 0:
            erosion_kernel = np.ones((3, 3), np.uint8)
            erode = cv2.erode(mask_array, erosion_kernel, iterations=self.erosion_iters)
            erode = np.where(erode > 0, 255, mask_array)
        else:
            erode = mask_array.copy()

        dilation = cv2.dilate(erode, kernel, iterations=1)

        dilation = np.where(dilation == 255, 127, dilation)  # WHITE to GRAY
        trimap = np.where(erode > 127, 200, dilation)  # mark the tumor inside GRAY

        trimap = np.where(trimap < 127, 0, trimap)  # Embelishment
        trimap = np.where(trimap > 200, 0, trimap)  # Embelishment
        trimap = np.where(trimap == 200, 255, trimap)  # GRAY to WHITE

        return PIL.Image.fromarray(trimap).convert("L")



import socket
import struct
from typing import Optional
from urllib.parse import urlparse


def is_loopback(address):
    host: Optional[str] = None

    try:
        parsed_url = urlparse(address)
        host = parsed_url.hostname
    except:
        return False  # url is not even a url

    loopback_checker = {
        socket.AF_INET: lambda x: struct.unpack('!I', socket.inet_aton(x))[0] >> (32 - 8) == 127,
        socket.AF_INET6: lambda x: x == '::1'
    }
    for family in (socket.AF_INET, socket.AF_INET6):
        try:
            r = socket.getaddrinfo(host, None, family, socket.SOCK_STREAM)
        except socket.gaierror:
            continue
        for family, _, _, _, sockaddr in r:
            if loopback_checker[family](sockaddr[0]):
                return True

    if host in ("localhost",):
        return True

    return False



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
from carvekit.ml.wrap.fba_matting import FBAMatting
from typing import Union, List
from PIL import Image
from pathlib import Path
from carvekit.trimap.cv_gen import CV2TrimapGenerator
from carvekit.trimap.generator import TrimapGenerator
from carvekit.utils.mask_utils import apply_mask
from carvekit.utils.pool_utils import thread_pool_processing
from carvekit.utils.image_utils import load_image, convert_image

__all__ = ["MattingMethod"]


class MattingMethod:
    """
    Improving the edges of the object mask using neural networks for matting and algorithms for creating trimap.
    Neural network for matting performs accurate object edge detection by using a special map called trimap,
    with unknown area that we scan for boundary, already known general object area and the background."""

    def __init__(self, matting_module: Union[FBAMatting],
                 trimap_generator: Union[TrimapGenerator, CV2TrimapGenerator],
                 device="cpu"):
        """
        Initializes Matting Method class.

        Args:
            matting_module: Initialized matting neural network class
            trimap_generator: Initialized trimap generator class
            device: Processing device used for applying mask to image
        """
        self.device = device
        self.matting_module = matting_module
        self.trimap_generator = trimap_generator

    def __call__(self, images: List[Union[str, Path, Image.Image]],
                 masks: List[Union[str, Path, Image.Image]]):
        """
        Passes data through apply_mask function

        Args:
            images: list of images
            masks: list pf masks

        Returns:
            list of images
        """
        if len(images) != len(masks):
            raise ValueError("Images and Masks lists should have same length!")
        images = thread_pool_processing(lambda x: convert_image(load_image(x)), images)
        masks = thread_pool_processing(lambda x: convert_image(load_image(x), mode="L"), masks)
        trimaps = thread_pool_processing(lambda x: self.trimap_generator(original_image=images[x],
                                                                         mask=masks[x]), range(len(images)))
        alpha = self.matting_module(images=images, trimaps=trimaps)
        return list(map(lambda x: apply_mask(image=images[x], mask=alpha[x], device=self.device),
                        range(len(images))))



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import numpy as np
from PIL import Image


def prob_filter(mask: Image.Image, prob_threshold=231) -> Image.Image:
    """
    Applies a filter to the mask by the probability of locating an object in the object area.

    Args:
        prob_threshold: Threshold of probability for mark area as background.
        mask: Predicted object mask

    Raises:
        ValueError if mask or trimap has wrong color mode

    Returns:
        Generated trimap for image.
    """
    if mask.mode != "L":
        raise ValueError("Input mask has wrong color mode.")
    # noinspection PyTypeChecker
    mask_array = np.array(mask)
    mask_array[mask_array > prob_threshold] = 255  # Probability filter for mask
    mask_array[mask_array <= prob_threshold] = 0
    return Image.fromarray(mask_array).convert("L")


def prob_as_unknown_area(trimap: Image.Image, mask: Image.Image, prob_threshold=255) -> Image.Image:
    """
    Marks any uncertainty in the seg mask as an unknown region.

    Args:
        prob_threshold: Threshold of probability for mark area as unknown.
        trimap: Generated trimap.
        mask: Predicted object mask

    Raises:
        ValueError if mask or trimap has wrong color mode

    Returns:
        Generated trimap for image.
    """
    if mask.mode != "L" or trimap.mode != "L":
        raise ValueError("Input mask has wrong color mode.")
    # noinspection PyTypeChecker
    mask_array = np.array(mask)
    # noinspection PyTypeChecker
    trimap_array = np.array(trimap)
    trimap_array[np.logical_and(mask_array <= prob_threshold, mask_array > 0)] = 127
    return Image.fromarray(trimap_array).convert("L")



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
from pathlib import Path
from PIL import Image
import warnings
from typing import Optional


def save_file(output: Optional[Path], input_path: Path, image: Image.Image):
    """
    Saves an image to the file system

    Args:
        output: Output path [dir or end file]
        input_path: Input path of the image
        image: Image to be saved.
    """
    if isinstance(output, Path) and str(output) != "none":
        if output.is_dir() and output.exists():
            image.save(output.joinpath(input_path.with_suffix('.png').name))
        elif output.suffix != '':
            if output.suffix != ".png":
                warnings.warn(f"Only export with .png extension is supported! Your {output.suffix}"
                              f" extension will be ignored and replaced with .png!")
            image.save(output.with_suffix('.png'))
        else:
            raise ValueError("Wrong output path!")
    elif output is None or str(output) == "none":
        image.save(input_path.with_name(input_path.stem.split('.')[0] + '_bg_removed').with_suffix('.png'))



"""
Modified by Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
Source url: https://github.com/MarcoForte/FBA_Matting
License: MIT License
"""
import torch
import torch.nn as nn
from torch.nn import functional as F


class Conv2d(nn.Conv2d):

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super(Conv2d, self).__init__(in_channels, out_channels, kernel_size, stride,
                                     padding, dilation, groups, bias)

    def forward(self, x):
        # return super(Conv2d, self).forward(x)
        weight = self.weight
        weight_mean = weight.mean(dim=1, keepdim=True).mean(dim=2,
                                                            keepdim=True).mean(dim=3, keepdim=True)
        weight = weight - weight_mean
        # std = (weight).view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1) + 1e-5
        std = torch.sqrt(torch.var(weight.view(weight.size(0), -1), dim=1) + 1e-12).view(-1, 1, 1, 1) + 1e-5
        weight = weight / std.expand_as(weight)
        return F.conv2d(x, weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)


def BatchNorm2d(num_features):
    return nn.GroupNorm(num_channels=num_features, num_groups=32)



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""

import random
import warnings
import torch


def fix_seed(seed=42):
    """Sets fixed random seed

    Args:
        seed: Random seed to be set
    """
    random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        # noinspection PyUnresolvedReferences
        torch.backends.cudnn.deterministic = True
        # noinspection PyUnresolvedReferences
        torch.backends.cudnn.benchmark = False
    return True


def suppress_warnings():
    # Suppress PyTorch 1.11.0 warning associated with changing order of args in nn.MaxPool2d layer,
    # since source code is not affected by this issue and there aren't any other correct way to hide this message.
    warnings.filterwarnings("ignore",
                            category=UserWarning,
                            message="Note that order of the arguments: ceil_mode and "
                                    "return_indices will changeto match the args list "
                                    "in nn.MaxPool2d in a future release.",
                            module="torch")



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
from pathlib import Path
from typing import Union, List

from PIL import Image

__all__ = ["PreprocessingStub"]




class PreprocessingStub:
    """Stub for future preprocessing methods"""

    def __call__(self, interface, images: List[Union[str, Path, Image.Image]]):
        """
        Passes data though interface.segmentation_pipeline() method

        Args:
            interface: Interface instance
            images: list of images

        Returns:
            the result of passing data through segmentation_pipeline method of interface
        """
        return interface.segmentation_pipeline(images=images)



def error_dict(error_text: str):
    """
    Generates a dictionary containing $error_text error
    :param error_text: Error text
    :return: error dictionary
    """
    resp = {"errors": [{"title": error_text}]}
    return resp



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
from pathlib import Path
from typing import Union, List, Optional

from PIL import Image

from carvekit.ml.wrap.basnet import BASNET
from carvekit.ml.wrap.deeplab_v3 import DeepLabV3
from carvekit.ml.wrap.u2net import U2NET
from carvekit.pipelines.preprocessing import PreprocessingStub
from carvekit.pipelines.postprocessing import MattingMethod
from carvekit.utils.image_utils import load_image
from carvekit.utils.mask_utils import apply_mask
from carvekit.utils.pool_utils import thread_pool_processing


class Interface:
    def __init__(self,
                 seg_pipe: Union[U2NET, BASNET, DeepLabV3],
                 pre_pipe: Optional[Union[PreprocessingStub]] = None,
                 post_pipe: Optional[Union[MattingMethod]] = None,
                 device="cpu"):
        """
        Initializes an object for interacting with pipelines and other components of the CarveKit framework.

        Args:
            pre_pipe: Initialized pre-processing pipeline object
            seg_pipe: Initialized segmentation network object
            post_pipe: Initialized postprocessing pipeline object
            device: The processing device that will be used to apply the masks to the images.
        """
        self.device = device
        self.preprocessing_pipeline = pre_pipe
        self.segmentation_pipeline = seg_pipe
        self.postprocessing_pipeline = post_pipe

    def __call__(self, images: List[Union[str, Path, Image.Image]]) -> List[Image.Image]:
        """
        Removes the background from the specified images.

        Args:
            images: list of input images

        Returns:
            List of images without background as PIL.Image.Image instances
        """
        images = thread_pool_processing(load_image, images)
        if self.preprocessing_pipeline is not None:
            masks: List[Image.Image] = self.preprocessing_pipeline(interface=self, images=images)
        else:
            masks: List[Image.Image] = self.segmentation_pipeline(images=images)

        if self.postprocessing_pipeline is not None:
            images: List[Image.Image] = self.postprocessing_pipeline(images=images, masks=masks)
        else:
            images = list(map(lambda x: apply_mask(image=images[x], mask=masks[x], device=self.device),
                              range(len(images))))
        return images



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
from PIL import Image
from carvekit.trimap.cv_gen import CV2TrimapGenerator
from carvekit.trimap.add_ops import prob_filter, prob_as_unknown_area


class TrimapGenerator(CV2TrimapGenerator):
    def __init__(self, prob_threshold: float = 231, kernel_size: int = 30, erosion_iters: int = 5):
        """
        Initialize a TrimapGenerator instance

        Args:
            prob_threshold: Probability threshold at which the
            prob_filter and prob_as_unknown_area operations will be applied
            kernel_size: The size of the offset from the object mask
            in pixels when an unknown area is detected in the trimap
            erosion_iters: The number of iterations of erosion that
            the object's mask will be subjected to before forming an unknown area
        """
        super().__init__(kernel_size, erosion_iters)
        self.prob_threshold = prob_threshold

    def __call__(self, original_image: Image.Image, mask: Image.Image) -> Image.Image:
        """
        Generates trimap based on predicted object mask to refine object mask borders.
        Based on cv2 erosion algorithm and additional prob. filters.
        Args:
            original_image: Original image
            mask: Predicted object mask

        Returns:
            Generated trimap for image.
        """
        filter_mask = prob_filter(mask=mask, prob_threshold=self.prob_threshold)
        trimap = super(TrimapGenerator, self).__call__(original_image, filter_mask)
        new_trimap = prob_as_unknown_area(trimap=trimap, mask=mask, prob_threshold=self.prob_threshold)

        return new_trimap



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import os
import re

from setuptools import setup, find_packages

from carvekit import version

IS_COLAB_PACKAGE = os.getenv("COLAB_PACKAGE_RELEASE", None)


def read(filename: str):
    filepath = os.path.join(os.path.dirname(__file__), filename)
    file = open(filepath, 'r', encoding='utf-8')
    return file.read()


def req_file(filename: str, folder: str = "."):
    with open(os.path.join(folder, filename), encoding='utf-8') as f:
        content = f.readlines()
    # you may also want to remove whitespace characters
    # Example: `\n` at the end of each line
    if os.getenv("COLAB_PACKAGE_RELEASE") is not None:
        return [re.sub("(~=.*)|(==.*)|(typing.*)", "", x.strip()) for x in content]
    return [x.strip() for x in content]


setup(
    name='carvekit' if IS_COLAB_PACKAGE is None else 'carvekit_colab',
    version=version,
    author="Nikita Selin (Anodev)",
    author_email='farvard34@gmail.com',
    description='Open-Source background removal framework',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    license='Apache License v2.0',
    keywords=["ml", "carvekit", "background removal", "neural networks", "machine learning", "remove bg"],
    url='https://github.com/OPHoperHPO/image-background-remove-tool',
    packages=find_packages(),
    scripts=[],
    install_requires=req_file("requirements.txt"),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'carvekit=carvekit:__main__.removebg',
        ],
    },
    python_requires=">=3.8" if IS_COLAB_PACKAGE is None else ">=3.6",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
)



import base64
import http
import io
import time
from json import JSONDecodeError
from typing import Optional

import requests
from PIL import Image
from fastapi import Header, Depends, Form, File, Request, APIRouter, UploadFile
from fastapi.openapi.models import Response
from pydantic import ValidationError
from starlette.responses import JSONResponse

from carvekit.web.deps import config, ml_processor
from carvekit.web.handlers.response import handle_response, Authenticate
from carvekit.web.responses.api import error_dict
from carvekit.web.schemas.request import Parameters
from carvekit.web.utils.net_utils import is_loopback

api_router = APIRouter(prefix='', tags=['api'])


# noinspection PyBroadException
@api_router.post('/removebg')
async def removebg(
        request: Request,
        image_file: Optional[bytes] = File(None),
        auth: bool = Depends(Authenticate),
        content_type: str = Header(""),
        image_file_b64: Optional[str] = Form(None),
        image_url: Optional[str] = Form(None),
        bg_image_file: Optional[bytes] = File(None),
        size: Optional[str] = Form("full"),
        type: Optional[str] = Form("auto"),
        format: Optional[str] = Form("auto"),
        roi: str = Form("0% 0% 100% 100%"),
        crop: bool = Form(False),
        crop_margin: Optional[str] = Form("0px"),
        scale: Optional[str] = Form("original"),
        position: Optional[str] = Form("original"),
        channels: Optional[str] = Form("rgba"),
        add_shadow: bool = Form(False),  # Not supported at the moment
        semitransparency: bool = Form(False),  # Not supported at the moment
        bg_color: Optional[str] = Form("")
):
    if auth is False:
        return JSONResponse(content=error_dict("Missing API Key"), status_code=403)
    if content_type not in ["application/x-www-form-urlencoded",
                            "application/json"] and "multipart/form-data" not in content_type:
        return JSONResponse(content=error_dict("Invalid request content type"), status_code=400)

    if image_url:
        if (
            not image_url.startswith("http://") or
            not image_url.startswith("https://") or
            is_loopback(image_url)
        ):
            print(f"Possible ssrf attempt to /api/removebg endpoint with image url: {image_url}")
            return JSONResponse(content=error_dict("Invalid image url."),
                                status_code=400)  # possible ssrf attempt

    image = None
    bg = None
    parameters = None
    if content_type == "application/x-www-form-urlencoded" or "multipart/form-data" in content_type:
        if image_file_b64 is None and image_url is None and image_file is None:
            return JSONResponse(content=error_dict("File not found"), status_code=400)

        if image_file_b64:
            if len(image_file_b64) == 0:
                return JSONResponse(content=error_dict("Empty image"), status_code=400)
            try:
                image = Image.open(io.BytesIO(base64.b64decode(image_file_b64)))
            except BaseException:
                return JSONResponse(content=error_dict("Error decode image!"), status_code=400)
        elif image_url:
            try:
                image = Image.open(io.BytesIO(requests.get(image_url).content))
            except BaseException:
                return JSONResponse(content=error_dict("Error download image!"), status_code=400)
        elif image_file:
            if len(image_file) == 0:
                return JSONResponse(content=error_dict("Empty image"), status_code=400)
            image = Image.open(io.BytesIO(image_file))

        if bg_image_file:
            if len(bg_image_file) == 0:
                return JSONResponse(content=error_dict("Empty image"), status_code=400)
            bg = Image.open(io.BytesIO(bg_image_file))
        try:
            parameters = Parameters(
                image_file_b64=image_file_b64,
                image_url=image_url,
                size=size,
                type=type,
                format=format,
                roi=roi,
                crop=crop,
                crop_margin=crop_margin,
                scale=scale,
                position=position,
                channels=channels,
                add_shadow=add_shadow,
                semitransparency=semitransparency,
                bg_color=bg_color,
            )
        except ValidationError as e:
            return JSONResponse(content=e.json(), status_code=400, media_type='application/json')

    else:
        payload = None
        try:
            payload = await request.json()
        except JSONDecodeError:
            return JSONResponse(content=error_dict("Empty json"), status_code=400)
        try:
            parameters = Parameters(**payload)
        except ValidationError as e:
            return Response(content=e.json(), status_code=400, media_type='application/json')
        if parameters.image_file_b64 is None and parameters.image_url is None:
            return JSONResponse(content=error_dict("File not found"), status_code=400)

        if parameters.image_file_b64:
            if len(parameters.image_file_b64) == 0:
                return JSONResponse(content=error_dict("Empty image"), status_code=400)
            try:
                image = Image.open(io.BytesIO(base64.b64decode(parameters.image_file_b64)))
            except BaseException:
                return JSONResponse(content=error_dict("Error decode image!"), status_code=400)
        elif parameters.image_url:
            if (
                    not parameters.image_url.startswith("http://") or
                    not parameters.image_url.startswith("https://") or
                    is_loopback(parameters.image_url)
            ):
                print(f"Possible ssrf attempt to /api/removebg endpoint with image url: {parameters.image_url}")
                return JSONResponse(content=error_dict("Invalid image url."),
                                    status_code=400)  # possible ssrf attempt
            try:
                image = Image.open(io.BytesIO(requests.get(parameters.image_url).content))
            except BaseException:
                return JSONResponse(content=error_dict("Error download image!"), status_code=400)
        if image is None:
            return JSONResponse(content=error_dict("Error download image!"), status_code=400)

    job_id = ml_processor.job_create([parameters.dict(), image, bg, False])

    while ml_processor.job_status(job_id) != "finished":
        if ml_processor.job_status(job_id) == "not_found":
            return JSONResponse(content=error_dict("Job ID not found!"), status_code=500)
        time.sleep(5)

    result = ml_processor.job_result(job_id)
    return handle_response(result, image)


@api_router.get("/account")
def account():
    """
    Stub for compatibility with remove.bg api libraries
    """
    return JSONResponse(content={"data": {"attributes": {
        "credits": {"total": 99999, "subscription": 99999, "payg": 99999, "enterprise": 99999},
        "api": {"free_calls": 99999, "sizes": "all"}}}}, status_code=200)


@api_router.get("/admin/config")
def status(auth: str = Depends(Authenticate)):
    """
    Returns the current server config.
    """
    if not auth or auth != "admin":
        return JSONResponse(content=error_dict("Authentication failed"), status_code=403)
    resp = JSONResponse(content=config.json(), status_code=200)
    resp.headers["X-Credits-Charged"] = "0"
    return resp



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
import hashlib
import os
import warnings
from pathlib import Path

import requests
import tqdm

MODELS_URLS = {
    "basnet.pth":
        "https://huggingface.co/anodev/basnet-universal/resolve/870becbdb364fda6d8fdb2c10b072542f8d08701/basnet.pth",
    "deeplab.pth":
        "https://huggingface.co/anodev/deeplabv3-resnet101/resolve/d504005392fc877565afdf58aad0cd524682d2b0/deeplab.pth",
    "fba_matting.pth":
        "https://huggingface.co/anodev/fba/resolve/a5d3457df0fb9c88ea19ed700d409756ca2069d1/fba_matting.pth",
    "u2net.pth":
        "https://huggingface.co/anodev/u2net-universal/resolve/10305d785481cf4b2eee1d447c39cd6e5f43d74b/full_weights"
        ".pth",
}

MODELS_CHECKSUMS = {
    "basnet.pth": "e409cb709f4abca87cb11bd44a9ad3f909044a917977ab65244b4c94dd33"
                  "8b1a37755c4253d7cb54526b7763622a094d7b676d34b5e6886689256754e5a5e6ad",
    "deeplab.pth":
        "9c5a1795bc8baa267200a44b49ac544a1ba2687d210f63777e4bd715387324469a59b072f8a28"
        "9cc471c637b367932177e5b312e8ea6351c1763d9ff44b4857c",
    "fba_matting.pth":
        "890906ec94c1bfd2ad08707a63e4ccb0955d7f5d25e32853950c24c78"
        "4cbad2e59be277999defc3754905d0f15aa75702cdead3cfe669ff72f08811c52971613",
    "u2net.pth":
        "16f8125e2fedd8c85db0e001ee15338b4aa2fda77bab8ba70c25e"
        "bea1533fda5ee70a909b934a9bd495b432cef89d629f00a07858a517742476fa8b346de24f7",

}


def download_model(path: Path) -> Path:
    """ Downloads model from repo.

    Args:
        path (pathlib.Path): Path to file

    Returns:
         Path if exists

    Raises:
        FileNotFoundError: if model checkpoint is not exists in known checkpoints models
        ConnectionError: if the model cannot be loaded from the URL.
    """
    if path.name in MODELS_URLS:
        model_url = MODELS_URLS[path.name]
        path.parent.mkdir(parents=True, exist_ok=True)
        try:
            r = requests.get(model_url, stream=True)
            if r.status_code == 200:
                with path.absolute().open('wb') as f:
                    r.raw.decode_content = True
                    for chunk in tqdm.tqdm(r, desc="Downloading " + path.name + ' model', colour='blue'):
                        f.write(chunk)
        except BaseException as e:
            if path.exists():
                os.remove(path)
            raise ConnectionError(f"Exception caused when downloading model! "
                                  f"Model name: {path.name}. Exception: {str(e)}")
        return path
    else:
        raise FileNotFoundError("Unknown model!")


def sha512_checksum_calc(file: Path) -> str:
    """
    Calculates the SHA512 hash digest of a file on fs

    Args:
        file: Path to the file

    Returns:
        SHA512 hash digest of a file.
    """
    dd = hashlib.sha512()
    with file.open("rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            dd.update(chunk)
    return dd.hexdigest()


def check_model(path: Path) -> bool:
    """ Verifies model checksums and existence in the file system

    Args:
        path: Path to the model

    Returns:
        True if all is okay and False if not

    Raises:
        FileNotFoundError: if model checkpoint is not exists in known checkpoints models
    """
    if path.exists():
        if path.name in MODELS_URLS:
            if MODELS_CHECKSUMS[path.name] != sha512_checksum_calc(path):
                warnings.warn(f"Invalid checksum for model {path.name}. Downloading correct model!")
                os.remove(path)
                return False
            return True
        else:
            raise FileNotFoundError("Unknown model!")
    else:
        return False


def check_for_exists(path: Path) -> Path:
    """ Checks for checkpoint path exists

    Args:
        path (pathlib.Path): Path to file

    Returns:
         Path if exists

    Raises:
        FileNotFoundError: if model checkpoint is not exists in known checkpoints models
        ConnectionError: if the model cannot be loaded from the URL.
    """
    if not check_model(path):
        download_model(path)

    return path



"""
Source url: https://github.com/OPHoperHPO/image-background-remove-tool
Author: Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
License: Apache License 2.0
"""
from carvekit.api.interface import Interface
from carvekit.ml.wrap.fba_matting import FBAMatting
from carvekit.ml.wrap.u2net import U2NET
from carvekit.pipelines.postprocessing import MattingMethod
from carvekit.trimap.generator import TrimapGenerator


class HiInterface(Interface):
    def __init__(self, batch_size_seg=5, batch_size_matting=1,
                 device='cpu', seg_mask_size=320, matting_mask_size=2048):
        """
        Initializes High Level interface.

        Args:
            matting_mask_size:  The size of the input image for the matting neural network.
            seg_mask_size: The size of the input image for the segmentation neural network.
            batch_size_seg: Number of images processed per one segmentation neural network call.
            batch_size_matting: Number of images processed per one matting neural network call.
            device: Processing device

        Notes:
            Changing seg_mask_size may cause an out-of-memory error if the value is too large, and it may also
            result in reduced precision. I do not recommend changing this value. You can change matting_mask_size in
            range from (1024 to 4096) to improve object edge refining quality, but it will cause extra large RAM and
            video memory consume. Also, you can change batch size to accelerate background removal, but it also causes
            extra large video memory consume, if value is too big.
        """
        self.u2net = U2NET(device=device, batch_size=batch_size_seg, input_image_size=seg_mask_size)
        self.fba = FBAMatting(batch_size=batch_size_matting, device=device, input_tensor_size=matting_mask_size)
        self.trimap_generator = TrimapGenerator()
        super(HiInterface, self).__init__(pre_pipe=None,
                                          seg_pipe=self.u2net,
                                          post_pipe=MattingMethod(matting_module=self.fba,
                                                                  trimap_generator=self.trimap_generator,
                                                                  device=device),
                                          device=device)


from os import getenv
from typing import Union

from loguru import logger

from carvekit.web.schemas.config import WebAPIConfig, MLConfig, AuthConfig
from carvekit.api.interface import Interface
from carvekit.ml.wrap.fba_matting import FBAMatting
from carvekit.ml.wrap.u2net import U2NET
from carvekit.ml.wrap.deeplab_v3 import DeepLabV3
from carvekit.ml.wrap.basnet import BASNET

from carvekit.pipelines.postprocessing import MattingMethod
from carvekit.pipelines.preprocessing import PreprocessingStub
from carvekit.trimap.generator import TrimapGenerator


def init_config() -> WebAPIConfig:
    default_config = WebAPIConfig()
    config = WebAPIConfig(
        **dict(port=int(getenv('CARVEKIT_PORT', default_config.port)),
               host=getenv('CARVEKIT_HOST', default_config.host),
               ml=MLConfig(
                   segmentation_network=getenv('CARVEKIT_SEGMENTATION_NETWORK', default_config.ml.segmentation_network),
                   preprocessing_method=getenv('CARVEKIT_PREPROCESSING_METHOD', default_config.ml.preprocessing_method),
                   postprocessing_method=getenv('CARVEKIT_POSTPROCESSING_METHOD',
                                                default_config.ml.postprocessing_method),
                   device=getenv('CARVEKIT_DEVICE', default_config.ml.device),
                   batch_size_seg=int(getenv('CARVEKIT_BATCH_SIZE_SEG', default_config.ml.batch_size_seg)),
                   batch_size_matting=int(getenv('CARVEKIT_BATCH_SIZE_MATTING', default_config.ml.batch_size_matting)),
                   seg_mask_size=int(getenv('CARVEKIT_SEG_MASK_SIZE', default_config.ml.seg_mask_size)),
                   matting_mask_size=int(getenv('CARVEKIT_MATTING_MASK_SIZE', default_config.ml.matting_mask_size))
               ), auth=AuthConfig(
                auth=bool(int(getenv('CARVEKIT_AUTH_ENABLE', default_config.auth.auth))),
                admin_token=getenv('CARVEKIT_ADMIN_TOKEN', default_config.auth.admin_token),
                allowed_tokens=default_config.auth.allowed_tokens
                if getenv('CARVEKIT_ALLOWED_TOKENS') is None else getenv('CARVEKIT_ALLOWED_TOKENS').split(',')

            ))

    )

    logger.info(f'Admin token for Web API is {config.auth.admin_token}')
    logger.debug(f"Running Web API with this config: {config.json()}")
    return config


def init_interface(config: Union[WebAPIConfig, MLConfig]) -> Interface:
    if isinstance(config, WebAPIConfig):
        config = config.ml
    if config.segmentation_network == "u2net":
        seg_net = U2NET(device=config.device,
                        batch_size=config.batch_size_seg,
                        input_image_size=config.seg_mask_size)
    elif config.segmentation_network == "deeplabv3":
        seg_net = DeepLabV3(device=config.device,
                            batch_size=config.batch_size_seg,
                            input_image_size=config.seg_mask_size)
    elif config.segmentation_network == "basnet":
        seg_net = BASNET(device=config.device,
                         batch_size=config.batch_size_seg,
                         input_tensor_size=config.seg_mask_size)
    else:
        seg_net = U2NET(device=config.device,
                        batch_size=config.batch_size_seg,
                        input_image_size=config.seg_mask_size)

    if config.preprocessing_method == "stub":
        preprocessing = PreprocessingStub()
    elif config.preprocessing_method == "none":
        preprocessing = None
    else:
        preprocessing = None

    if config.postprocessing_method == "fba":
        fba = FBAMatting(device=config.device,
                         batch_size=config.batch_size_matting,
                         input_tensor_size=config.matting_mask_size)
        trimap_generator = TrimapGenerator()
        postprocessing = MattingMethod(device=config.device,
                                       matting_module=fba,
                                       trimap_generator=trimap_generator)

    elif config.postprocessing_method == "none":
        postprocessing = None
    else:
        postprocessing = None

    interface = Interface(pre_pipe=preprocessing,
                          post_pipe=postprocessing,
                          seg_pipe=seg_net,
                          device=config.device)
    return interface



"""
Modified by Nikita Selin (OPHoperHPO)[https://github.com/OPHoperHPO].
Source url: https://github.com/xuebinqin/U-2-Net
License: Apache License 2.0
"""
from typing import Union

import torch
import torch.nn as nn

import math

__all__ = ['U2NETArchitecture']


def _upsample_like(x, size):
    return nn.Upsample(size=size, mode='bilinear', align_corners=False)(x)


def _size_map(x, height):
    # {height: size} for Upsample
    size = list(x.shape[-2:])
    sizes = {}
    for h in range(1, height):
        sizes[h] = size
        size = [math.ceil(w / 2) for w in size]
    return sizes


class REBNCONV(nn.Module):
    def __init__(self, in_ch=3, out_ch=3, dilate=1):
        super(REBNCONV, self).__init__()

        self.conv_s1 = nn.Conv2d(in_ch, out_ch, 3, padding=1 * dilate, dilation=1 * dilate)
        self.bn_s1 = nn.BatchNorm2d(out_ch)
        self.relu_s1 = nn.ReLU(inplace=True)

    def forward(self, x):
        return self.relu_s1(self.bn_s1(self.conv_s1(x)))


class RSU(nn.Module):
    def __init__(self, name, height, in_ch, mid_ch, out_ch, dilated=False):
        super(RSU, self).__init__()
        self.name = name
        self.height = height
        self.dilated = dilated
        self._make_layers(height, in_ch, mid_ch, out_ch, dilated)

    def forward(self, x):
        sizes = _size_map(x, self.height)
        x = self.rebnconvin(x)

        # U-Net like symmetric encoder-decoder structure
        def unet(x, height=1):
            if height < self.height:
                x1 = getattr(self, f'rebnconv{height}')(x)
                if not self.dilated and height < self.height - 1:
                    x2 = unet(getattr(self, 'downsample')(x1), height + 1)
                else:
                    x2 = unet(x1, height + 1)

                x = getattr(self, f'rebnconv{height}d')(torch.cat((x2, x1), 1))
                return _upsample_like(x, sizes[height - 1]) if not self.dilated and height > 1 else x
            else:
                return getattr(self, f'rebnconv{height}')(x)

        return x + unet(x)

    def _make_layers(self, height, in_ch, mid_ch, out_ch, dilated=False):
        self.add_module('rebnconvin', REBNCONV(in_ch, out_ch))
        self.add_module('downsample', nn.MaxPool2d(2, stride=2, ceil_mode=True))

        self.add_module(f'rebnconv1', REBNCONV(out_ch, mid_ch))
        self.add_module(f'rebnconv1d', REBNCONV(mid_ch * 2, out_ch))

        for i in range(2, height):
            dilate = 1 if not dilated else 2 ** (i - 1)
            self.add_module(f'rebnconv{i}', REBNCONV(mid_ch, mid_ch, dilate=dilate))
            self.add_module(f'rebnconv{i}d', REBNCONV(mid_ch * 2, mid_ch, dilate=dilate))

        dilate = 2 if not dilated else 2 ** (height - 1)
        self.add_module(f'rebnconv{height}', REBNCONV(mid_ch, mid_ch, dilate=dilate))


class U2NETArchitecture(nn.Module):
    def __init__(self, cfg_type: Union[dict, str] = "full", out_ch: int = 1):
        super(U2NETArchitecture, self).__init__()
        if isinstance(cfg_type, str):
            if cfg_type == "full":
                layers_cfgs = {
                    # cfgs for building RSUs and sides
                    # {stage : [name, (height(L), in_ch, mid_ch, out_ch, dilated), side]}
                    'stage1': ['En_1', (7, 3, 32, 64), -1],
                    'stage2': ['En_2', (6, 64, 32, 128), -1],
                    'stage3': ['En_3', (5, 128, 64, 256), -1],
                    'stage4': ['En_4', (4, 256, 128, 512), -1],
                    'stage5': ['En_5', (4, 512, 256, 512, True), -1],
                    'stage6': ['En_6', (4, 512, 256, 512, True), 512],
                    'stage5d': ['De_5', (4, 1024, 256, 512, True), 512],
                    'stage4d': ['De_4', (4, 1024, 128, 256), 256],
                    'stage3d': ['De_3', (5, 512, 64, 128), 128],
                    'stage2d': ['De_2', (6, 256, 32, 64), 64],
                    'stage1d': ['De_1', (7, 128, 16, 64), 64],
                }
            else:
                raise ValueError("Unknown U^2-Net architecture conf. name")
        elif isinstance(cfg_type, dict):
            layers_cfgs = cfg_type
        else:
            raise ValueError("Unknown U^2-Net architecture conf. type")
        self.out_ch = out_ch
        self._make_layers(layers_cfgs)

    def forward(self, x):
        sizes = _size_map(x, self.height)
        maps = []  # storage for maps

        # side saliency map
        def unet(x, height=1):
            if height < 6:
                x1 = getattr(self, f'stage{height}')(x)
                x2 = unet(getattr(self, 'downsample')(x1), height + 1)
                x = getattr(self, f'stage{height}d')(torch.cat((x2, x1), 1))
                side(x, height)
                return _upsample_like(x, sizes[height - 1]) if height > 1 else x
            else:
                x = getattr(self, f'stage{height}')(x)
                side(x, height)
                return _upsample_like(x, sizes[height - 1])

        def side(x, h):
            # side output saliency map (before sigmoid)
            x = getattr(self, f'side{h}')(x)
            x = _upsample_like(x, sizes[1])
            maps.append(x)

        def fuse():
            # fuse saliency probability maps
            maps.reverse()
            x = torch.cat(maps, 1)
            x = getattr(self, 'outconv')(x)
            maps.insert(0, x)
            return [torch.sigmoid(x) for x in maps]

        unet(x)
        maps = fuse()
        return maps

    def _make_layers(self, cfgs):
        self.height = int((len(cfgs) + 1) / 2)
        self.add_module('downsample', nn.MaxPool2d(2, stride=2, ceil_mode=True))
        for k, v in cfgs.items():
            # build rsu block
            self.add_module(k, RSU(v[0], *v[1]))
            if v[2] > 0:
                # build side layer
                self.add_module(f'side{v[0][-1]}', nn.Conv2d(v[2], self.out_ch, 3, padding=1))
        # build fuse layer
        self.add_module('outconv', nn.Conv2d(int(self.height * self.out_ch), self.out_ch, 1))



import re
from typing import Optional

from pydantic import BaseModel, validator
from typing_extensions import Literal


class Parameters(BaseModel):
    image_file_b64: Optional[str] = ""
    image_url: Optional[str] = ""
    size: Optional[Literal['preview', 'full', 'auto']] = "preview"
    type: Optional[Literal['auto', 'product', 'person', 'car']] = "auto"  # Not supported at the moment
    format: Optional[Literal['auto', 'jpg', 'png', 'zip']] = "auto"
    roi: str = "0% 0% 100% 100%"
    crop: bool = False
    crop_margin: Optional[str] = "0px"
    scale: Optional[str] = "original"
    position: Optional[str] = "original"
    channels: Optional[Literal['rgba', 'alpha']] = "rgba"
    add_shadow: str = "false"  # Not supported at the moment
    semitransparency: str = "false"  # Not supported at the moment
    bg_color: Optional[str] = ""
    bg_image_url: Optional[str] = ""

    @validator('crop_margin')
    def crop_margin_validator(cls, value):
        if not re.match(r'[0-9]+(px|%)$', value):
            raise ValueError('crop_margin paramter is not valid')  # TODO: Add support of several values
        if '%' in value and (int(value[:-1]) < 0 or int(value[:-1]) > 100):
            raise ValueError('crop_margin mast be in range between 0% and 100%')
        return value

    @validator('scale')
    def scale_validator(cls, value):
        if value != 'original' and (
                not re.match(r'[0-9]+%$', value) or not int(value[:-1]) <= 100 or not int(value[:-1]) >= 10):
            raise ValueError('scale must be original or in between of 10% and 100%')

        if value == 'original':
            return 100

        return int(value[:-1])

    @validator('position')
    def position_validator(cls, value, values):
        if len(value.split(' ')) > 2:
            raise ValueError(
                'Position must be a value from 0 to 100 '
                'for both vertical and horizontal axises or for both axises respectively')

        if value == 'original':
            return 'original'
        elif len(value.split(' ')) == 1:
            return [int(value[:-1]), int(value[:-1])]
        else:
            return [int(value.split(' ')[0][:-1]), int(value.split(' ')[1][:-1])]

    @validator('bg_color')
    def bg_color_validator(cls, value):
        if not re.match(r'(#{0,1}[0-9a-f]{3}){0,2}$', value):
            raise ValueError('bg_color is not in hex')
        if len(value) and value[0] != '#':
            value = '#' + value
        return value
